# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# pylint: skip-file


from __future__ import absolute_import, division, print_function

from salt.ext.tornado.httputil import HTTPHeaders, HTTPMessageDelegate, HTTPServerConnectionDelegate, ResponseStartLine
from salt.ext.tornado.routing import HostMatches, PathMatches, ReversibleRouter, Router, Rule, RuleRouter
from salt.ext.tornado.testing import AsyncHTTPTestCase
from salt.ext.tornado.web import Application, HTTPError, RequestHandler
from salt.ext.tornado.wsgi import WSGIContainer


class BasicRouter(Router):
    def find_handler(self, request, **kwargs):

        class MessageDelegate(HTTPMessageDelegate):
            def __init__(self, connection):
                self.connection = connection

            def finish(self):
                self.connection.write_headers(
                    ResponseStartLine("HTTP/1.1", 200, "OK"), HTTPHeaders({"Content-Length": "2"}), b"OK"
                )
                self.connection.finish()

        return MessageDelegate(request.connection)


class BasicRouterTestCase(AsyncHTTPTestCase):
    def get_app(self):
        return BasicRouter()

    def test_basic_router(self):
        response = self.fetch("/any_request")
        self.assertEqual(response.body, b"OK")


resources = {}


class GetResource(RequestHandler):
    def get(self, path):
        if path not in resources:
            raise HTTPError(404)

        self.finish(resources[path])


class PostResource(RequestHandler):
    def post(self, path):
        resources[path] = self.request.body


class HTTPMethodRouter(Router):
    def __init__(self, app):
        self.app = app

    def find_handler(self, request, **kwargs):
        handler = GetResource if request.method == "GET" else PostResource
        return self.app.get_handler_delegate(request, handler, path_args=[request.path])


class HTTPMethodRouterTestCase(AsyncHTTPTestCase):
    def get_app(self):
        return HTTPMethodRouter(Application())

    def test_http_method_router(self):
        response = self.fetch("/post_resource", method="POST", body="data")
        self.assertEqual(response.code, 200)

        response = self.fetch("/get_resource")
        self.assertEqual(response.code, 404)

        response = self.fetch("/post_resource")
        self.assertEqual(response.code, 200)
        self.assertEqual(response.body, b"data")


def _get_named_handler(handler_name):
    class Handler(RequestHandler):
        def get(self, *args, **kwargs):
            if self.application.settings.get("app_name") is not None:
                self.write(self.application.settings["app_name"] + ": ")

            self.finish(handler_name + ": " + self.reverse_url(handler_name))

    return Handler


FirstHandler = _get_named_handler("first_handler")
SecondHandler = _get_named_handler("second_handler")


class CustomRouter(ReversibleRouter):
    def __init__(self):
        super(CustomRouter, self).__init__()
        self.routes = {}

    def add_routes(self, routes):
        self.routes.update(routes)

    def find_handler(self, request, **kwargs):
        if request.path in self.routes:
            app, handler = self.routes[request.path]
            return app.get_handler_delegate(request, handler)

    def reverse_url(self, name, *args):
        handler_path = '/' + name
        return handler_path if handler_path in self.routes else None


class CustomRouterTestCase(AsyncHTTPTestCase):
    def get_app(self):
        class CustomApplication(Application):
            def reverse_url(self, name, *args):
                return router.reverse_url(name, *args)

        router = CustomRouter()
        app1 = CustomApplication(app_name="app1")
        app2 = CustomApplication(app_name="app2")

        router.add_routes({
            "/first_handler": (app1, FirstHandler),
            "/second_handler": (app2, SecondHandler),
            "/first_handler_second_app": (app2, FirstHandler),
        })

        return router

    def test_custom_router(self):
        response = self.fetch("/first_handler")
        self.assertEqual(response.body, b"app1: first_handler: /first_handler")
        response = self.fetch("/second_handler")
        self.assertEqual(response.body, b"app2: second_handler: /second_handler")
        response = self.fetch("/first_handler_second_app")
        self.assertEqual(response.body, b"app2: first_handler: /first_handler")


class ConnectionDelegate(HTTPServerConnectionDelegate):
    def start_request(self, server_conn, request_conn):

        class MessageDelegate(HTTPMessageDelegate):
            def __init__(self, connection):
                self.connection = connection

            def finish(self):
                response_body = b"OK"
                self.connection.write_headers(
                    ResponseStartLine("HTTP/1.1", 200, "OK"),
                    HTTPHeaders({"Content-Length": str(len(response_body))}))
                self.connection.write(response_body)
                self.connection.finish()

        return MessageDelegate(request_conn)


class RuleRouterTest(AsyncHTTPTestCase):
    def get_app(self):
        app = Application()

        def request_callable(request):
            request.write(b"HTTP/1.1 200 OK\r\nContent-Length: 2\r\n\r\nOK")
            request.finish()

        app.add_handlers(".*", [
            (HostMatches("www.example.com"), [
                (PathMatches("/first_handler"), "tornado.test.routing_test.SecondHandler", {}, "second_handler")
            ]),
            Rule(PathMatches("/first_handler"), FirstHandler, name="first_handler"),
            Rule(PathMatches("/request_callable"), request_callable),
            ("/connection_delegate", ConnectionDelegate())
        ])

        return app

    def test_rule_based_router(self):
        response = self.fetch("/first_handler")
        self.assertEqual(response.body, b"first_handler: /first_handler")
        response = self.fetch("/first_handler", headers={'Host': 'www.example.com'})
        self.assertEqual(response.body, b"second_handler: /first_handler")

        response = self.fetch("/connection_delegate")
        self.assertEqual(response.body, b"OK")

        response = self.fetch("/request_callable")
        self.assertEqual(response.body, b"OK")

        response = self.fetch("/404")
        self.assertEqual(response.code, 404)


class WSGIContainerTestCase(AsyncHTTPTestCase):
    def get_app(self):
        wsgi_app = WSGIContainer(self.wsgi_app)

        class Handler(RequestHandler):
            def get(self, *args, **kwargs):
                self.finish(self.reverse_url("tornado"))

        return RuleRouter([
            (PathMatches("/tornado.*"), Application([(r"/tornado/test", Handler, {}, "tornado")])),
            (PathMatches("/wsgi"), wsgi_app),
        ])

    def wsgi_app(self, environ, start_response):
        start_response("200 OK", [])
        return [b"WSGI"]

    def test_wsgi_container(self):
        response = self.fetch("/tornado/test")
        self.assertEqual(response.body, b"/tornado/test")

        response = self.fetch("/wsgi")
        self.assertEqual(response.body, b"WSGI")



# This software released into the public domain. Anyone is free to copy,
# modify, publish, use, compile, sell, or distribute this software,
# either in source code form or as a compiled binary, for any purpose,
# commercial or non-commercial, and by any means.

import socket
import ctypes
import os
from salt._compat import ipaddress


class sockaddr(ctypes.Structure):
    _fields_ = [("sa_family", ctypes.c_short),
                ("__pad1", ctypes.c_ushort),
                ("ipv4_addr", ctypes.c_byte * 4),
                ("ipv6_addr", ctypes.c_byte * 16),
                ("__pad2", ctypes.c_ulong)]


if hasattr(ctypes, 'windll'):
    WSAStringToAddressA = ctypes.windll.ws2_32.WSAStringToAddressA
    WSAAddressToStringA = ctypes.windll.ws2_32.WSAAddressToStringA
else:
    def not_windows(*args):
        raise SystemError(
            "Invalid platform. ctypes.windll must be available."
        )
    WSAStringToAddressA = not_windows
    WSAAddressToStringA = not_windows


def inet_pton(address_family, ip_string):
    # Verify IP Address
    # This will catch IP Addresses such as 10.1.2
    if address_family == socket.AF_INET:
        try:
            ipaddress.ip_address(str(ip_string))
        except ValueError:
            raise OSError('illegal IP address string passed to inet_pton')
        return socket.inet_aton(ip_string)

    # Verify IP Address
    # The `WSAStringToAddressA` function handles notations used by Berkeley
    # software which includes 3 part IP Addresses such as `10.1.2`. That's why
    # the above check is needed to enforce more strict IP Address validation as
    # used by the `inet_pton` function in Unix.
    # See the following:
    # https://stackoverflow.com/a/29286098
    # Docs for the `inet_addr` function on MSDN
    # https://msdn.microsoft.com/en-us/library/windows/desktop/ms738563.aspx
    addr = sockaddr()
    addr.sa_family = address_family
    addr_size = ctypes.c_int(ctypes.sizeof(addr))

    if WSAStringToAddressA(
            ip_string.encode('ascii'),
            address_family,
            None,
            ctypes.byref(addr),
            ctypes.byref(addr_size)
    ) != 0:
        raise OSError(ctypes.FormatError())

    if address_family == socket.AF_INET:
        return ctypes.string_at(addr.ipv4_addr, 4)
    if address_family == socket.AF_INET6:
        return ctypes.string_at(addr.ipv6_addr, 16)

    raise OSError('unknown address family')


def inet_ntop(address_family, packed_ip):
    addr = sockaddr()
    addr.sa_family = address_family
    addr_size = ctypes.c_int(ctypes.sizeof(addr))
    ip_string = ctypes.create_string_buffer(128)
    ip_string_size = ctypes.c_int(ctypes.sizeof(ip_string))

    if address_family == socket.AF_INET:
        if len(packed_ip) != ctypes.sizeof(addr.ipv4_addr):
            raise OSError('packed IP wrong length for inet_ntoa')
        ctypes.memmove(addr.ipv4_addr, packed_ip, 4)
    elif address_family == socket.AF_INET6:
        if len(packed_ip) != ctypes.sizeof(addr.ipv6_addr):
            raise OSError('packed IP wrong length for inet_ntoa')
        ctypes.memmove(addr.ipv6_addr, packed_ip, 16)
    else:
        raise OSError('unknown address family')

    if WSAAddressToStringA(
            ctypes.byref(addr),
            addr_size,
            None,
            ip_string,
            ctypes.byref(ip_string_size)
    ) != 0:
        raise OSError(ctypes.FormatError())

    return ip_string[:ip_string_size.value - 1]


# Adding our two functions to the socket library
if os.name == 'nt':
    socket.inet_pton = inet_pton
    socket.inet_ntop = inet_ntop



import urllib.parse

import pytest

import salt.utils.json


@pytest.mark.slow_test
@pytest.mark.netapi_client_data(["local", "runner"])
async def test_accepts_arg_kwarg_keys(
    http_client, auth_creds, content_type_map, subtests
):
    """
    Ensure that (singular) arg and kwarg keys (for passing parameters)
    are supported by runners.
    """
    # Login to get the token
    body = salt.utils.json.dumps(auth_creds)
    response = await http_client.fetch(
        "/login",
        method="POST",
        body=body,
        headers={
            "Accept": content_type_map["json"],
            "Content-Type": content_type_map["json"],
        },
    )
    assert response.code == 200
    token = response.headers["X-Auth-Token"]
    low1 = {
        "client": "runner",
        "fun": "test.arg",
        "arg": [1234, 5678],
        "kwarg": {"ext_source": "redis"},
    }
    low2 = {
        "client": "runner",
        "fun": "test.arg",
        "arg": 1234,
        "kwarg": {"ext_source": "redis"},
    }
    for content_type in ("json", "form"):
        with subtests.test(content_type=content_type):
            if content_type == "json":
                body = salt.utils.json.dumps(low1)
            else:
                _low = low1.copy()
                arg = _low.pop("arg")
                body = urllib.parse.urlencode(_low)
                for _arg in arg:
                    body += "&arg={}".format(_arg)
            response = await http_client.fetch(
                "/",
                method="POST",
                body=body,
                headers={
                    "Accept": content_type_map["json"],
                    "Content-Type": content_type_map[content_type],
                    "X-Auth-Token": token,
                },
            )
            assert response.code == 200
            body = salt.utils.json.loads(response.body)
            ret = body["return"][0]
            assert ret["args"] == low1["arg"]
            assert ret["kwargs"] == low1["kwarg"]
            if content_type == "json":
                body = salt.utils.json.dumps(low2)
            else:
                _low = low2.copy()
                arg = _low.pop("arg")
                body = urllib.parse.urlencode(_low)
                body += "&arg={}".format(arg)
            response = await http_client.fetch(
                "/",
                method="POST",
                body=body,
                headers={
                    "Accept": content_type_map["json"],
                    "Content-Type": content_type_map[content_type],
                    "X-Auth-Token": token,
                },
            )
            assert response.code == 200
            body = salt.utils.json.loads(response.body)
            ret = body["return"][0]
            assert ret["args"][0] == low2["arg"]
            assert ret["kwargs"] == low1["kwarg"]



import salt.utils.parsers
from salt.utils.verify import check_user


class SaltKey(salt.utils.parsers.SaltKeyOptionParser):
    """
    Initialize the Salt key manager
    """

    def run(self):
        """
        Execute salt-key
        """
        import salt.key

        self.parse_args()

        key = salt.key.KeyCLI(self.config)
        if check_user(self.config["user"]):
            key.run()



"""
    :codeauthor: `Gareth J. Greenaway <gareth@saltstack.com>`

    tests.unit.cloud.clouds.digitalocean_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import logging

import pytest

from salt.cloud.clouds import digitalocean
from salt.exceptions import SaltCloudSystemExit

log = logging.getLogger(__name__)


def test_reboot_no_call():
    """
    Tests that a SaltCloudSystemExit is raised when
    kwargs that are provided do not include an action.
    """
    with pytest.raises(SaltCloudSystemExit) as excinfo:
        digitalocean.reboot(name="fake_name")

    assert "The reboot action must be called with -a or --action." == str(excinfo.value)



"""
    :codeauthor: :email:`Jakub Sliva <jakub.sliva@ultimum.io>`
"""

import pytest

import salt.states.zabbix_valuemap as zabbix_valuemap
from tests.support.mock import MagicMock, patch


@pytest.fixture
def input_params():
    return {
        "mappings": [
            {"newvalue": "OK", "value": "0h"},
            {"newvalue": "Failure", "value": "1"},
        ],
        "name": "Server HP Health",
    }


@pytest.fixture
def existing_obj():
    return [
        {
            "valuemapid": "21",
            "name": "Server HP Health",
            "mappings": [
                {"newvalue": "OK", "value": "0h"},
                {"newvalue": "Failure", "value": "1"},
            ],
        }
    ]


@pytest.fixture
def existing_obj_diff():
    return {
        "valuemapid": "21",
        "name": "Server HP Health",
        "mappings": [
            {"newvalue": "OK", "value": "0h"},
            {"newvalue": "Failure", "value": "1"},
            {"newvalue": "some", "value": "2"},
        ],
    }


@pytest.fixture
def diff_params():
    return {
        "valuemapid": "21",
        "mappings": [
            {"newvalue": "OK", "value": "0h"},
            {"newvalue": "Failure", "value": "1"},
        ],
    }


@pytest.fixture
def configure_loader_modules():
    return {zabbix_valuemap: {}}


def test_present_create(input_params):
    """
    Test to ensure that named value map is created
    """
    name = "Server HP Health"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    def side_effect_run_query(*args):
        """
        Differentiate between __salt__ exec module function calls with different parameters.
        """
        if args[0] == "valuemap.get":
            return False
        elif args[0] == "valuemap.create":
            return True

    with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
        with patch.dict(
            zabbix_valuemap.__salt__,
            {
                "zabbix.get_zabbix_id_mapper": MagicMock(
                    return_value={"valuemap": "valuemapid"}
                ),
                "zabbix.substitute_params": MagicMock(
                    side_effect=[input_params, False]
                ),
                "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                "zabbix.compare_params": MagicMock(return_value={}),
            },
        ):
            ret["result"] = True
            ret["comment"] = 'Zabbix Value map "{}" created.'.format(name)
            ret["changes"] = {
                name: {
                    "old": 'Zabbix Value map "{}" did not exist.'.format(name),
                    "new": 'Zabbix Value map "{}" created according definition.'.format(
                        name
                    ),
                }
            }
            assert zabbix_valuemap.present(name, {}) == ret


def test_present_exists(input_params, existing_obj):
    """
    Test to ensure that named value map is present and not changed
    """
    name = "Server HP Health"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
        with patch.dict(
            zabbix_valuemap.__salt__,
            {
                "zabbix.get_zabbix_id_mapper": MagicMock(
                    return_value={"valuemap": "valuemapid"}
                ),
                "zabbix.substitute_params": MagicMock(
                    side_effect=[input_params, existing_obj]
                ),
                "zabbix.run_query": MagicMock(return_value=["length of result is 1"]),
                "zabbix.compare_params": MagicMock(return_value={}),
            },
        ):
            ret["result"] = True
            ret[
                "comment"
            ] = 'Zabbix Value map "{}" already exists and corresponds to a definition.'.format(
                name
            )
            assert zabbix_valuemap.present(name, {}) == ret


def test_present_update(input_params, existing_obj_diff, diff_params):
    """
    Test to ensure that named value map is present but must be updated
    """
    name = "Server HP Health"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    def side_effect_run_query(*args):
        """
        Differentiate between __salt__ exec module function calls with different parameters.
        """
        if args[0] == "valuemap.get":
            return ["length of result is 1 = valuemap exists"]
        elif args[0] == "valuemap.update":
            return diff_params

    with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
        with patch.dict(
            zabbix_valuemap.__salt__,
            {
                "zabbix.get_zabbix_id_mapper": MagicMock(
                    return_value={"valuemap": "valuemapid"}
                ),
                "zabbix.substitute_params": MagicMock(
                    side_effect=[input_params, existing_obj_diff]
                ),
                "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                "zabbix.compare_params": MagicMock(return_value=diff_params),
            },
        ):
            ret["result"] = True
            ret["comment"] = 'Zabbix Value map "{}" updated.'.format(name)
            ret["changes"] = {
                name: {
                    "old": (
                        'Zabbix Value map "{}" differed '
                        "in following parameters: {}".format(name, diff_params)
                    ),
                    "new": 'Zabbix Value map "{}" fixed.'.format(name),
                }
            }
            assert zabbix_valuemap.present(name, {}) == ret


def test_absent_test_mode():
    """
    Test to ensure that named value map is absent in test mode
    """
    name = "Server HP Health"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    with patch.dict(zabbix_valuemap.__opts__, {"test": True}):
        with patch.dict(
            zabbix_valuemap.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
        ):
            ret["result"] = True
            ret["comment"] = 'Zabbix Value map "{}" would be deleted.'.format(name)
            ret["changes"] = {
                name: {
                    "old": 'Zabbix Value map "{}" exists.'.format(name),
                    "new": 'Zabbix Value map "{}" would be deleted.'.format(name),
                }
            }
            assert zabbix_valuemap.absent(name) == ret


def test_absent():
    """
    Test to ensure that named value map is absent
    """
    name = "Server HP Health"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
        with patch.dict(
            zabbix_valuemap.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=False)},
        ):
            ret["result"] = True
            ret["comment"] = 'Zabbix Value map "{}" does not exist.'.format(name)
            assert zabbix_valuemap.absent(name) == ret

        with patch.dict(
            zabbix_valuemap.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
        ):
            with patch.dict(
                zabbix_valuemap.__salt__,
                {"zabbix.run_query": MagicMock(return_value=True)},
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{}" deleted.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Value map "{}" existed.'.format(name),
                        "new": 'Zabbix Value map "{}" deleted.'.format(name),
                    }
                }
                assert zabbix_valuemap.absent(name) == ret



import logging
import random
import string

import pytest

import salt.config
import salt.loader
import salt.modules.boto_cloudtrail as boto_cloudtrail
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

# pylint: disable=import-error,no-name-in-module,unused-import
try:
    import boto
    import boto3
    from botocore.exceptions import ClientError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

# pylint: enable=import-error,no-name-in-module,unused-import

# the boto_cloudtrail module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = "1.2.1"

log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto3.__version__) < Version(required_boto3_version):
        return False
    else:
        return True


if _has_required_boto():
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    not_found_error = ClientError(
        {"Error": {"Code": "TrailNotFoundException", "Message": "Test-defined error"}},
        "msg",
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    trail_ret = dict(
        Name="testtrail",
        IncludeGlobalServiceEvents=True,
        KmsKeyId=None,
        LogFileValidationEnabled=False,
        S3BucketName="auditinfo",
        TrailARN="arn:aws:cloudtrail:us-east-1:214351231622:trail/testtrail",
    )
    status_ret = dict(
        IsLogging=False,
        LatestCloudWatchLogsDeliveryError=None,
        LatestCloudWatchLogsDeliveryTime=None,
        LatestDeliveryError=None,
        LatestDeliveryTime=None,
        LatestDigestDeliveryError=None,
        LatestDigestDeliveryTime=None,
        LatestNotificationError=None,
        LatestNotificationTime=None,
        StartLoggingTime=None,
        StopLoggingTime=None,
    )


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoCloudTrailTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto3", "args", "systemd", "path", "platform"], context={}
        )
        return {boto_cloudtrail: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto_cloudtrail.__init__(self.opts)
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        session_instance.client.return_value = self.conn


class BotoCloudTrailTestCaseMixin:
    pass


class BotoCloudTrailTestCase(BotoCloudTrailTestCaseBase, BotoCloudTrailTestCaseMixin):
    """
    TestCase for salt.modules.boto_cloudtrail module
    """

    def test_that_when_checking_if_a_trail_exists_and_a_trail_exists_the_trail_exists_method_returns_true(
        self,
    ):
        """
        Tests checking cloudtrail trail existence when the cloudtrail trail already exists
        """
        self.conn.get_trail_status.return_value = trail_ret
        result = boto_cloudtrail.exists(Name=trail_ret["Name"], **conn_parameters)

        self.assertTrue(result["exists"])

    def test_that_when_checking_if_a_trail_exists_and_a_trail_does_not_exist_the_trail_exists_method_returns_false(
        self,
    ):
        """
        Tests checking cloudtrail trail existence when the cloudtrail trail does not exist
        """
        self.conn.get_trail_status.side_effect = not_found_error
        result = boto_cloudtrail.exists(Name="mytrail", **conn_parameters)

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_a_trail_exists_and_boto3_returns_an_error_the_trail_exists_method_returns_error(
        self,
    ):
        """
        Tests checking cloudtrail trail existence when boto returns an error
        """
        self.conn.get_trail_status.side_effect = ClientError(
            error_content, "get_trail_status"
        )
        result = boto_cloudtrail.exists(Name="mytrail", **conn_parameters)

        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("get_trail_status"),
        )

    def test_that_when_creating_a_trail_succeeds_the_create_trail_method_returns_true(
        self,
    ):
        """
        tests True trail created.
        """
        self.conn.create_trail.return_value = trail_ret
        result = boto_cloudtrail.create(
            Name=trail_ret["Name"],
            S3BucketName=trail_ret["S3BucketName"],
            **conn_parameters
        )

        self.assertTrue(result["created"])

    def test_that_when_creating_a_trail_fails_the_create_trail_method_returns_error(
        self,
    ):
        """
        tests False trail not created.
        """
        self.conn.create_trail.side_effect = ClientError(error_content, "create_trail")
        result = boto_cloudtrail.create(
            Name=trail_ret["Name"],
            S3BucketName=trail_ret["S3BucketName"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("create_trail")
        )

    def test_that_when_deleting_a_trail_succeeds_the_delete_trail_method_returns_true(
        self,
    ):
        """
        tests True trail deleted.
        """
        result = boto_cloudtrail.delete(Name="testtrail", **conn_parameters)

        self.assertTrue(result["deleted"])

    def test_that_when_deleting_a_trail_fails_the_delete_trail_method_returns_false(
        self,
    ):
        """
        tests False trail not deleted.
        """
        self.conn.delete_trail.side_effect = ClientError(error_content, "delete_trail")
        result = boto_cloudtrail.delete(Name="testtrail", **conn_parameters)
        self.assertFalse(result["deleted"])

    def test_that_when_describing_trail_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if trail exists
        """
        self.conn.describe_trails.return_value = {"trailList": [trail_ret]}

        result = boto_cloudtrail.describe(Name=trail_ret["Name"], **conn_parameters)

        self.assertTrue(result["trail"])

    def test_that_when_describing_trail_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters if trail does not exist
        """
        self.conn.describe_trails.side_effect = not_found_error
        result = boto_cloudtrail.describe(Name="testtrail", **conn_parameters)

        self.assertFalse(result["trail"])

    def test_that_when_describing_trail_on_client_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        self.conn.describe_trails.side_effect = ClientError(error_content, "get_trail")
        result = boto_cloudtrail.describe(Name="testtrail", **conn_parameters)
        self.assertTrue("error" in result)

    def test_that_when_getting_status_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests getting status if trail exists
        """
        self.conn.get_trail_status.return_value = status_ret

        result = boto_cloudtrail.status(Name=trail_ret["Name"], **conn_parameters)

        self.assertTrue(result["trail"])

    def test_that_when_getting_status_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests getting status if trail does not exist
        """
        self.conn.get_trail_status.side_effect = not_found_error
        result = boto_cloudtrail.status(Name="testtrail", **conn_parameters)

        self.assertFalse(result["trail"])

    def test_that_when_getting_status_on_client_error_it_returns_error(self):
        """
        Tests getting status failure
        """
        self.conn.get_trail_status.side_effect = ClientError(
            error_content, "get_trail_status"
        )
        result = boto_cloudtrail.status(Name="testtrail", **conn_parameters)
        self.assertTrue("error" in result)

    def test_that_when_listing_trails_succeeds_the_list_trails_method_returns_true(
        self,
    ):
        """
        tests True trails listed.
        """
        self.conn.describe_trails.return_value = {"trailList": [trail_ret]}
        result = boto_cloudtrail.list(**conn_parameters)

        self.assertTrue(result["trails"])

    def test_that_when_listing_trail_fails_the_list_trail_method_returns_false(self):
        """
        tests False no trail listed.
        """
        self.conn.describe_trails.return_value = {"trailList": []}
        result = boto_cloudtrail.list(**conn_parameters)
        self.assertFalse(result["trails"])

    def test_that_when_listing_trail_fails_the_list_trail_method_returns_error(self):
        """
        tests False trail error.
        """
        self.conn.describe_trails.side_effect = ClientError(
            error_content, "list_trails"
        )
        result = boto_cloudtrail.list(**conn_parameters)
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("list_trails")
        )

    def test_that_when_updating_a_trail_succeeds_the_update_trail_method_returns_true(
        self,
    ):
        """
        tests True trail updated.
        """
        self.conn.update_trail.return_value = trail_ret
        result = boto_cloudtrail.update(
            Name=trail_ret["Name"],
            S3BucketName=trail_ret["S3BucketName"],
            **conn_parameters
        )

        self.assertTrue(result["updated"])

    def test_that_when_updating_a_trail_fails_the_update_trail_method_returns_error(
        self,
    ):
        """
        tests False trail not updated.
        """
        self.conn.update_trail.side_effect = ClientError(error_content, "update_trail")
        result = boto_cloudtrail.update(
            Name=trail_ret["Name"],
            S3BucketName=trail_ret["S3BucketName"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("update_trail")
        )

    def test_that_when_starting_logging_succeeds_the_start_logging_method_returns_true(
        self,
    ):
        """
        tests True logging started.
        """
        result = boto_cloudtrail.start_logging(
            Name=trail_ret["Name"], **conn_parameters
        )

        self.assertTrue(result["started"])

    def test_that_when_start_logging_fails_the_start_logging_method_returns_false(self):
        """
        tests False logging not started.
        """
        self.conn.describe_trails.return_value = {"trailList": []}
        self.conn.start_logging.side_effect = ClientError(
            error_content, "start_logging"
        )
        result = boto_cloudtrail.start_logging(
            Name=trail_ret["Name"], **conn_parameters
        )
        self.assertFalse(result["started"])

    def test_that_when_stopping_logging_succeeds_the_stop_logging_method_returns_true(
        self,
    ):
        """
        tests True logging stopped.
        """
        result = boto_cloudtrail.stop_logging(Name=trail_ret["Name"], **conn_parameters)

        self.assertTrue(result["stopped"])

    def test_that_when_stop_logging_fails_the_stop_logging_method_returns_false(self):
        """
        tests False logging not stopped.
        """
        self.conn.describe_trails.return_value = {"trailList": []}
        self.conn.stop_logging.side_effect = ClientError(error_content, "stop_logging")
        result = boto_cloudtrail.stop_logging(Name=trail_ret["Name"], **conn_parameters)
        self.assertFalse(result["stopped"])

    def test_that_when_adding_tags_succeeds_the_add_tags_method_returns_true(self):
        """
        tests True tags added.
        """
        with patch.dict(
            boto_cloudtrail.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_cloudtrail.add_tags(
                Name=trail_ret["Name"], a="b", **conn_parameters
            )

        self.assertTrue(result["tagged"])

    def test_that_when_adding_tags_fails_the_add_tags_method_returns_false(self):
        """
        tests False tags not added.
        """
        self.conn.add_tags.side_effect = ClientError(error_content, "add_tags")
        with patch.dict(
            boto_cloudtrail.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_cloudtrail.add_tags(
                Name=trail_ret["Name"], a="b", **conn_parameters
            )
        self.assertFalse(result["tagged"])

    def test_that_when_removing_tags_succeeds_the_remove_tags_method_returns_true(self):
        """
        tests True tags removed.
        """
        with patch.dict(
            boto_cloudtrail.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_cloudtrail.remove_tags(
                Name=trail_ret["Name"], a="b", **conn_parameters
            )

        self.assertTrue(result["tagged"])

    def test_that_when_removing_tags_fails_the_remove_tags_method_returns_false(self):
        """
        tests False tags not removed.
        """
        self.conn.remove_tags.side_effect = ClientError(error_content, "remove_tags")
        with patch.dict(
            boto_cloudtrail.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_cloudtrail.remove_tags(
                Name=trail_ret["Name"], a="b", **conn_parameters
            )
        self.assertFalse(result["tagged"])

    def test_that_when_listing_tags_succeeds_the_list_tags_method_returns_true(self):
        """
        tests True tags listed.
        """
        with patch.dict(
            boto_cloudtrail.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_cloudtrail.list_tags(
                Name=trail_ret["Name"], **conn_parameters
            )

        self.assertEqual(result["tags"], {})

    def test_that_when_listing_tags_fails_the_list_tags_method_returns_false(self):
        """
        tests False tags not listed.
        """
        self.conn.list_tags.side_effect = ClientError(error_content, "list_tags")
        with patch.dict(
            boto_cloudtrail.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_cloudtrail.list_tags(
                Name=trail_ret["Name"], **conn_parameters
            )
        self.assertTrue(result["error"])



"""
Test the core grains
"""


import pytest

from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
class TestGrainsCore(ModuleCase):
    """
    Test the core grains grains
    """

    @pytest.mark.slow_test
    def test_grains_passed_to_custom_grain(self):
        """
        test if current grains are passed to grains module functions that have a grains argument
        """
        self.assertEqual(
            self.run_function("grains.get", ["custom_grain_test"]), "itworked"
        )



import attr
import pytest

import salt.defaults.events


@attr.s(kw_only=True, frozen=True, slots=True)
class PillarRefreshComplete:
    event_listener = attr.ib()
    minion_id = attr.ib()

    def __call__(self, start_time):
        expected_tag = salt.defaults.events.MINION_PILLAR_REFRESH_COMPLETE
        event_pattern = (self.minion_id, expected_tag)
        matched_events = self.event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=120
        )
        assert matched_events.found_all_events


@pytest.fixture(scope="module")
def wait_for_pillar_refresh_complete(event_listener, salt_minion):
    return PillarRefreshComplete(
        event_listener=event_listener, minion_id=salt_minion.id
    )



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def file(states):
    return states.file


@pytest.fixture(scope="function")
def single_dir_with_file(tmp_path):
    file = tmp_path / "stuff.txt"
    file.write_text("things")
    yield str(tmp_path)


@pytest.fixture(scope="function")
def nested_empty_dirs(tmp_path):
    num_root = 2
    num_mid = 4
    num_last = 2
    for root in range(1, num_root + 1):
        for mid in range(1, num_mid + 1):
            for last in range(1, num_last + 1):
                nest = (
                    tmp_path
                    / "root{}".format(root)
                    / "mid{}".format(mid)
                    / "last{}".format(last)
                )
                nest.mkdir(parents=True, exist_ok=True)
    yield str(tmp_path)


@pytest.fixture(scope="function")
def nested_dirs_with_files(tmp_path):
    num_root = 2
    num_mid = 4
    num_last = 2
    for root in range(1, num_root + 1):
        for mid in range(1, num_mid + 1):
            for last in range(1, num_last + 1):
                nest = (
                    tmp_path
                    / "root{}".format(root)
                    / "mid{}".format(mid)
                    / "last{}".format(last)
                )
                nest.mkdir(parents=True, exist_ok=True)
                if last % 2:
                    last_file = nest / "stuff.txt"
                    last_file.write_text("things")
    yield str(tmp_path)


def test_pruned_failure(file, single_dir_with_file):
    ret = file.pruned(name=single_dir_with_file)
    assert ret.result is False
    assert not ret.changes["deleted"]
    assert len(ret.changes["errors"]) == 1
    assert ret.comment == "Failed to remove directory {}".format(single_dir_with_file)


def test_pruned_success_recurse_and_deleted(file, nested_empty_dirs):
    ret = file.pruned(name=nested_empty_dirs, recurse=True)
    assert ret.result is True
    assert len(ret.changes["deleted"]) == 27
    assert ret.comment == "Recursively removed empty directories under {}".format(
        nested_empty_dirs
    )


def test_pruned_success_ignore_errors_and_deleted(file, nested_dirs_with_files):
    ret = file.pruned(name=nested_dirs_with_files, ignore_errors=True)
    assert ret.result is True
    assert len(ret.changes["deleted"]) == 8
    assert ret.comment == "Recursively removed empty directories under {}".format(
        nested_dirs_with_files
    )



#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""WSGI support for the Tornado web framework.

WSGI is the Python standard for web servers, and allows for interoperability
between Tornado and other Python web frameworks and servers.  This module
provides WSGI support in two ways:

* `WSGIAdapter` converts a `tornado.web.Application` to the WSGI application
  interface.  This is useful for running a Tornado app on another
  HTTP server, such as Google App Engine.  See the `WSGIAdapter` class
  documentation for limitations that apply.
* `WSGIContainer` lets you run other WSGI applications and frameworks on the
  Tornado HTTP server.  For example, with this class you can mix Django
  and Tornado handlers in a single server.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import sys
from io import BytesIO
import salt.ext.tornado as tornado

from salt.ext.tornado.concurrent import Future
from salt.ext.tornado import escape
from salt.ext.tornado import httputil
from salt.ext.tornado.log import access_log
from salt.ext.tornado import web
from salt.ext.tornado.escape import native_str
from salt.ext.tornado.util import unicode_type, PY3


if PY3:
    import urllib.parse as urllib_parse  # py3
else:
    import urllib as urllib_parse

# PEP 3333 specifies that WSGI on python 3 generally deals with byte strings
# that are smuggled inside objects of type unicode (via the latin1 encoding).
# These functions are like those in the tornado.escape module, but defined
# here to minimize the temptation to use them in non-wsgi contexts.
if str is unicode_type:
    def to_wsgi_str(s):
        assert isinstance(s, bytes)
        return s.decode('latin1')

    def from_wsgi_str(s):
        assert isinstance(s, str)
        return s.encode('latin1')
else:
    def to_wsgi_str(s):
        assert isinstance(s, bytes)
        return s

    def from_wsgi_str(s):
        assert isinstance(s, str)
        return s


class WSGIApplication(web.Application):
    """A WSGI equivalent of `tornado.web.Application`.

    .. deprecated:: 4.0

       Use a regular `.Application` and wrap it in `WSGIAdapter` instead.
    """
    def __call__(self, environ, start_response):
        return WSGIAdapter(self)(environ, start_response)


# WSGI has no facilities for flow control, so just return an already-done
# Future when the interface requires it.
_dummy_future = Future()
_dummy_future.set_result(None)


class _WSGIConnection(httputil.HTTPConnection):
    def __init__(self, method, start_response, context):
        self.method = method
        self.start_response = start_response
        self.context = context
        self._write_buffer = []
        self._finished = False
        self._expected_content_remaining = None
        self._error = None

    def set_close_callback(self, callback):
        # WSGI has no facility for detecting a closed connection mid-request,
        # so we can simply ignore the callback.
        pass

    def write_headers(self, start_line, headers, chunk=None, callback=None):
        if self.method == 'HEAD':
            self._expected_content_remaining = 0
        elif 'Content-Length' in headers:
            self._expected_content_remaining = int(headers['Content-Length'])
        else:
            self._expected_content_remaining = None
        self.start_response(
            '%s %s' % (start_line.code, start_line.reason),
            [(native_str(k), native_str(v)) for (k, v) in headers.get_all()])
        if chunk is not None:
            self.write(chunk, callback)
        elif callback is not None:
            callback()
        return _dummy_future

    def write(self, chunk, callback=None):
        if self._expected_content_remaining is not None:
            self._expected_content_remaining -= len(chunk)
            if self._expected_content_remaining < 0:
                self._error = httputil.HTTPOutputError(
                    "Tried to write more data than Content-Length")
                raise self._error
        self._write_buffer.append(chunk)
        if callback is not None:
            callback()
        return _dummy_future

    def finish(self):
        if (self._expected_content_remaining is not None and
                self._expected_content_remaining != 0):
            self._error = httputil.HTTPOutputError(
                "Tried to write %d bytes less than Content-Length" %
                self._expected_content_remaining)
            raise self._error
        self._finished = True


class _WSGIRequestContext(object):
    def __init__(self, remote_ip, protocol):
        self.remote_ip = remote_ip
        self.protocol = protocol

    def __str__(self):
        return self.remote_ip


class WSGIAdapter(object):
    """Converts a `tornado.web.Application` instance into a WSGI application.

    Example usage::

        import tornado.web
        import tornado.wsgi
        import wsgiref.simple_server

        class MainHandler(tornado.web.RequestHandler):
            def get(self):
                self.write("Hello, world")

        if __name__ == "__main__":
            application = tornado.web.Application([
                (r"/", MainHandler),
            ])
            wsgi_app = tornado.wsgi.WSGIAdapter(application)
            server = wsgiref.simple_server.make_server('', 8888, wsgi_app)
            server.serve_forever()

    See the `appengine demo
    <https://github.com/tornadoweb/tornado/tree/stable/demos/appengine>`_
    for an example of using this module to run a Tornado app on Google
    App Engine.

    In WSGI mode asynchronous methods are not supported.  This means
    that it is not possible to use `.AsyncHTTPClient`, or the
    `tornado.auth` or `tornado.websocket` modules.

    .. versionadded:: 4.0
    """
    def __init__(self, application):
        if isinstance(application, WSGIApplication):
            self.application = lambda request: web.Application.__call__(
                application, request)
        else:
            self.application = application

    def __call__(self, environ, start_response):
        method = environ["REQUEST_METHOD"]
        uri = urllib_parse.quote(from_wsgi_str(environ.get("SCRIPT_NAME", "")))
        uri += urllib_parse.quote(from_wsgi_str(environ.get("PATH_INFO", "")))
        if environ.get("QUERY_STRING"):
            uri += "?" + environ["QUERY_STRING"]
        headers = httputil.HTTPHeaders()
        if environ.get("CONTENT_TYPE"):
            headers["Content-Type"] = environ["CONTENT_TYPE"]
        if environ.get("CONTENT_LENGTH"):
            headers["Content-Length"] = environ["CONTENT_LENGTH"]
        for key in environ:
            if key.startswith("HTTP_"):
                headers[key[5:].replace("_", "-")] = environ[key]
        if headers.get("Content-Length"):
            body = environ["wsgi.input"].read(
                int(headers["Content-Length"]))
        else:
            body = b""
        protocol = environ["wsgi.url_scheme"]
        remote_ip = environ.get("REMOTE_ADDR", "")
        if environ.get("HTTP_HOST"):
            host = environ["HTTP_HOST"]
        else:
            host = environ["SERVER_NAME"]
        connection = _WSGIConnection(method, start_response,
                                     _WSGIRequestContext(remote_ip, protocol))
        request = httputil.HTTPServerRequest(
            method, uri, "HTTP/1.1", headers=headers, body=body,
            host=host, connection=connection)
        request._parse_body()
        self.application(request)
        if connection._error:
            raise connection._error
        if not connection._finished:
            raise Exception("request did not finish synchronously")
        return connection._write_buffer


class WSGIContainer(object):
    r"""Makes a WSGI-compatible function runnable on Tornado's HTTP server.

    .. warning::

       WSGI is a *synchronous* interface, while Tornado's concurrency model
       is based on single-threaded asynchronous execution.  This means that
       running a WSGI app with Tornado's `WSGIContainer` is *less scalable*
       than running the same app in a multi-threaded WSGI server like
       ``gunicorn`` or ``uwsgi``.  Use `WSGIContainer` only when there are
       benefits to combining Tornado and WSGI in the same process that
       outweigh the reduced scalability.

    Wrap a WSGI function in a `WSGIContainer` and pass it to `.HTTPServer` to
    run it. For example::

        def simple_app(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-type", "text/plain")]
            start_response(status, response_headers)
            return ["Hello world!\n"]

        container = tornado.wsgi.WSGIContainer(simple_app)
        http_server = tornado.httpserver.HTTPServer(container)
        http_server.listen(8888)
        tornado.ioloop.IOLoop.current().start()

    This class is intended to let other frameworks (Django, web.py, etc)
    run on the Tornado HTTP server and I/O loop.

    The `tornado.web.FallbackHandler` class is often useful for mixing
    Tornado and WSGI apps in the same server.  See
    https://github.com/bdarnell/django-tornado-demo for a complete example.
    """
    def __init__(self, wsgi_application):
        self.wsgi_application = wsgi_application

    def __call__(self, request):
        data = {}
        response = []

        def start_response(status, response_headers, exc_info=None):
            data["status"] = status
            data["headers"] = response_headers
            return response.append
        app_response = self.wsgi_application(
            WSGIContainer.environ(request), start_response)
        try:
            response.extend(app_response)
            body = b"".join(response)
        finally:
            if hasattr(app_response, "close"):
                app_response.close()
        if not data:
            raise Exception("WSGI app did not call start_response")

        status_code, reason = data["status"].split(' ', 1)
        status_code = int(status_code)
        headers = data["headers"]
        header_set = set(k.lower() for (k, v) in headers)
        body = escape.utf8(body)
        if status_code != 304:
            if "content-length" not in header_set:
                headers.append(("Content-Length", str(len(body))))
            if "content-type" not in header_set:
                headers.append(("Content-Type", "text/html; charset=UTF-8"))
        if "server" not in header_set:
            headers.append(("Server", "TornadoServer/%s" % salt.ext.tornado.version))

        start_line = httputil.ResponseStartLine("HTTP/1.1", status_code, reason)
        header_obj = httputil.HTTPHeaders()
        for key, value in headers:
            header_obj.add(key, value)
        request.connection.write_headers(start_line, header_obj, chunk=body)
        request.connection.finish()
        self._log(status_code, request)

    @staticmethod
    def environ(request):
        """Converts a `tornado.httputil.HTTPServerRequest` to a WSGI environment.
        """
        hostport = request.host.split(":")
        if len(hostport) == 2:
            host = hostport[0]
            port = int(hostport[1])
        else:
            host = request.host
            port = 443 if request.protocol == "https" else 80
        environ = {
            "REQUEST_METHOD": request.method,
            "SCRIPT_NAME": "",
            "PATH_INFO": to_wsgi_str(escape.url_unescape(
                request.path, encoding=None, plus=False)),
            "QUERY_STRING": request.query,
            "REMOTE_ADDR": request.remote_ip,
            "SERVER_NAME": host,
            "SERVER_PORT": str(port),
            "SERVER_PROTOCOL": request.version,
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": request.protocol,
            "wsgi.input": BytesIO(escape.utf8(request.body)),
            "wsgi.errors": sys.stderr,
            "wsgi.multithread": False,
            "wsgi.multiprocess": True,
            "wsgi.run_once": False,
        }
        if "Content-Type" in request.headers:
            environ["CONTENT_TYPE"] = request.headers.pop("Content-Type")
        if "Content-Length" in request.headers:
            environ["CONTENT_LENGTH"] = request.headers.pop("Content-Length")
        for key, value in request.headers.items():
            environ["HTTP_" + key.replace("-", "_").upper()] = value
        return environ

    def _log(self, status_code, request):
        if status_code < 400:
            log_method = access_log.info
        elif status_code < 500:
            log_method = access_log.warning
        else:
            log_method = access_log.error
        request_time = 1000.0 * request.request_time()
        summary = request.method + " " + request.uri + " (" + \
            request.remote_ip + ")"
        log_method("%d %s %.2fms", status_code, summary, request_time)


HTTPRequest = httputil.HTTPServerRequest



"""
Module for retrieving random information from Random.org

.. versionadded:: 2015.5.0

:configuration: This module can be used by either passing an api key and version
    directly or by specifying both in a configuration profile in the salt
    master/minion config.

    For example:

    .. code-block:: yaml

        random_org:
          api_key: 7be1402d-5719-5bd3-a306-3def9f135da5
          api_version: 1
"""
import http.client
import logging
import urllib.request

import salt.utils.http
import salt.utils.json

log = logging.getLogger(__name__)
__virtualname__ = "random_org"

RANDOM_ORG_FUNCTIONS = {
    "1": {
        "getUsage": {"method": "getUsage"},
        "generateIntegers": {"method": "generateIntegers"},
        "generateStrings": {"method": "generateStrings"},
        "generateUUIDs": {"method": "generateUUIDs"},
        "generateDecimalFractions": {"method": "generateDecimalFractions"},
        "generateGaussians": {"method": "generateGaussians"},
        "generateBlobs": {"method": "generateBlobs"},
    }
}


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    return __virtualname__


def _numeric(n):
    """
    Tell whether an argument is numeric
    """
    return isinstance(n, (int, float))


def _query(api_version=None, data=None):
    """
    Slack object method function to construct and execute on the API URL.

    :param api_key:     The Random.org api key.
    :param api_version: The version of Random.org api.
    :param data:        The data to be sent for POST method.
    :return:            The json response from the API call or False.
    """

    if data is None:
        data = {}

    ret = {"res": True}

    api_url = "https://api.random.org/"
    base_url = urllib.parse.urljoin(api_url, "json-rpc/" + str(api_version) + "/invoke")

    data = salt.utils.json.dumps(data)

    result = salt.utils.http.query(
        base_url,
        method="POST",
        params={},
        data=data,
        decode=True,
        status=True,
        header_dict={},
        opts=__opts__,
    )

    if result.get("status", None) == http.client.OK:
        _result = result["dict"]
        if _result.get("result"):
            return _result.get("result")
        if _result.get("error"):
            return _result.get("error")
        return False
    elif result.get("status", None) == http.client.NO_CONTENT:
        return False
    else:
        ret["message"] = result.text if hasattr(result, "text") else ""
        return ret


def getUsage(api_key=None, api_version=None):
    """
    Show current usages statistics

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :return: The current usage statistics.

    CLI Example:

    .. code-block:: bash

        salt '*' random_org.getUsage

        salt '*' random_org.getUsage api_key=peWcBiMOS9HrZG15peWcBiMOS9HrZG15 api_version=1
    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    if isinstance(api_version, int):
        api_version = str(api_version)

    _function = RANDOM_ORG_FUNCTIONS.get(api_version).get("getUsage").get("method")
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {"apiKey": api_key}

    result = _query(api_version=api_version, data=data)

    if result:
        ret["bitsLeft"] = result.get("bitsLeft")
        ret["requestsLeft"] = result.get("requestsLeft")
        ret["totalBits"] = result.get("totalBits")
        ret["totalRequests"] = result.get("totalRequests")
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret


def generateIntegers(api_key=None, api_version=None, **kwargs):
    """
    Generate random integers

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :param number: The number of integers to generate
    :param minimum: The lower boundary for the range from which the
                    random numbers will be picked. Must be within
                    the [-1e9,1e9] range.
    :param maximum: The upper boundary for the range from which the
                    random numbers will be picked. Must be within
                    the [-1e9,1e9] range.
    :param replacement: Specifies whether the random numbers should
                        be picked with replacement. The default (true)
                        will cause the numbers to be picked with replacement,
                        i.e., the resulting numbers may contain duplicate
                        values (like a series of dice rolls). If you want the
                        numbers picked to be unique (like raffle tickets drawn
                        from a container), set this value to false.
    :param base: Specifies the base that will be used to display the numbers.
                 Values allowed are 2, 8, 10 and 16. This affects the JSON
                 types and formatting of the resulting data as discussed below.
    :return: A list of integers.

    CLI Example:

    .. code-block:: bash

        salt '*' random_org.generateIntegers number=5 minimum=1 maximum=6

        salt '*' random_org.generateIntegers number=5 minimum=2 maximum=255 base=2

    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    for item in ["number", "minimum", "maximum"]:
        if item not in kwargs:
            ret["res"] = False
            ret["message"] = "Rquired argument, {} is missing.".format(item)
            return ret

    if not _numeric(kwargs["number"]) or not 1 <= kwargs["number"] <= 10000:
        ret["res"] = False
        ret["message"] = "Number of integers must be between 1 and 10000"
        return ret

    if (
        not _numeric(kwargs["minimum"])
        or not -1000000000 <= kwargs["minimum"] <= 1000000000
    ):
        ret["res"] = False
        ret[
            "message"
        ] = "Minimum argument must be between -1,000,000,000 and 1,000,000,000"
        return ret

    if (
        not _numeric(kwargs["maximum"])
        or not -1000000000 <= kwargs["maximum"] <= 1000000000
    ):
        ret["res"] = False
        ret[
            "message"
        ] = "Maximum argument must be between -1,000,000,000 and 1,000,000,000"
        return ret

    if "base" in kwargs:
        base = kwargs["base"]
        if base not in [2, 8, 10, 16]:
            ret["res"] = False
            ret["message"] = "Base must be either 2, 8, 10 or 16."
            return ret
    else:
        base = 10

    if "replacement" not in kwargs:
        replacement = True
    else:
        replacement = kwargs["replacement"]

    if isinstance(api_version, int):
        api_version = str(api_version)

    _function = (
        RANDOM_ORG_FUNCTIONS.get(api_version).get("generateIntegers").get("method")
    )
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {
        "apiKey": api_key,
        "n": kwargs["number"],
        "min": kwargs["minimum"],
        "max": kwargs["maximum"],
        "replacement": replacement,
        "base": base,
    }

    result = _query(api_version=api_version, data=data)
    log.debug("result %s", result)
    if result:
        if "random" in result:
            random_data = result.get("random").get("data")
            ret["data"] = random_data
        else:
            ret["res"] = False
            ret["message"] = result["message"]
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret


def generateStrings(api_key=None, api_version=None, **kwargs):
    """
    Generate random strings.

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :param number: The number of strings to generate.
    :param length: The length of each string. Must be
                   within the [1,20] range. All strings
                   will be of the same length
    :param characters: A string that contains the set of
                       characters that are allowed to occur
                       in the random strings. The maximum number
                       of characters is 80.
    :param replacement: Specifies whether the random strings should be picked
                        with replacement. The default (true) will cause the
                        strings to be picked with replacement, i.e., the
                        resulting list of strings may contain duplicates (like
                        a series of dice rolls). If you want the strings to be
                        unique (like raffle tickets drawn from a container), set
                        this value to false.
    :return: A list of strings.

    CLI Example:

    .. code-block:: bash

        salt '*' random_org.generateStrings number=5 length=8 characters='abcdefghijklmnopqrstuvwxyz'

        salt '*' random_org.generateStrings number=10 length=16 characters'abcdefghijklmnopqrstuvwxyz'

    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    for item in ["number", "length", "characters"]:
        if item not in kwargs:
            ret["res"] = False
            ret["message"] = "Required argument, {} is missing.".format(item)
            return ret

    if not _numeric(kwargs["number"]) or not 1 <= kwargs["number"] <= 10000:
        ret["res"] = False
        ret["message"] = "Number of strings must be between 1 and 10000"
        return ret

    if not _numeric(kwargs["length"]) or not 1 <= kwargs["length"] <= 20:
        ret["res"] = False
        ret["message"] = "Length of strings must be between 1 and 20"
        return ret

    if len(kwargs["characters"]) >= 80:
        ret["res"] = False
        ret["message"] = "Length of characters must be less than 80."
        return ret

    if isinstance(api_version, int):
        api_version = str(api_version)

    if "replacement" not in kwargs:
        replacement = True
    else:
        replacement = kwargs["replacement"]

    _function = (
        RANDOM_ORG_FUNCTIONS.get(api_version).get("generateStrings").get("method")
    )
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {
        "apiKey": api_key,
        "n": kwargs["number"],
        "length": kwargs["length"],
        "characters": kwargs["characters"],
        "replacement": replacement,
    }

    result = _query(api_version=api_version, data=data)
    if result:
        if "random" in result:
            random_data = result.get("random").get("data")
            ret["data"] = random_data
        else:
            ret["res"] = False
            ret["message"] = result["message"]
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret


def generateUUIDs(api_key=None, api_version=None, **kwargs):
    """
    Generate a list of random UUIDs

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :param number: How many random UUIDs you need.
                   Must be within the [1,1e3] range.
    :return: A list of UUIDs

    CLI Example:

    .. code-block:: bash

        salt '*' random_org.generateUUIDs number=5

    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    for item in ["number"]:
        if item not in kwargs:
            ret["res"] = False
            ret["message"] = "Required argument, {} is missing.".format(item)
            return ret

    if isinstance(api_version, int):
        api_version = str(api_version)

    if not _numeric(kwargs["number"]) or not 1 <= kwargs["number"] <= 1000:
        ret["res"] = False
        ret["message"] = "Number of UUIDs must be between 1 and 1000"
        return ret

    _function = RANDOM_ORG_FUNCTIONS.get(api_version).get("generateUUIDs").get("method")
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {
        "apiKey": api_key,
        "n": kwargs["number"],
    }

    result = _query(api_version=api_version, data=data)
    if result:
        if "random" in result:
            random_data = result.get("random").get("data")
            ret["data"] = random_data
        else:
            ret["res"] = False
            ret["message"] = result["message"]
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret


def generateDecimalFractions(api_key=None, api_version=None, **kwargs):
    """
    Generates true random decimal fractions

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :param number: How many random decimal fractions
                   you need. Must be within the [1,1e4] range.
    :param decimalPlaces: The number of decimal places
                          to use. Must be within the [1,20] range.
    :param replacement: Specifies whether the random numbers should
                        be picked with replacement. The default (true)
                        will cause the numbers to be picked with replacement,
                        i.e., the resulting numbers may contain duplicate
                        values (like a series of dice rolls). If you want the
                        numbers picked to be unique (like raffle tickets drawn
                        from a container), set this value to false.
    :return: A list of decimal fraction

    CLI Example:

    .. code-block:: bash

        salt '*' random_org.generateDecimalFractions number=10 decimalPlaces=4

        salt '*' random_org.generateDecimalFractions number=10 decimalPlaces=4 replacement=True

    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    for item in ["number", "decimalPlaces"]:
        if item not in kwargs:
            ret["res"] = False
            ret["message"] = "Required argument, {} is missing.".format(item)
            return ret

    if not isinstance(kwargs["number"], int) or not 1 <= kwargs["number"] <= 10000:
        ret["res"] = False
        ret["message"] = "Number of decimal fractions must be between 1 and 10000"
        return ret

    if not _numeric(kwargs["decimalPlaces"]) or not 1 <= kwargs["decimalPlaces"] <= 20:
        ret["res"] = False
        ret["message"] = "Number of decimal places must be between 1 and 20"
        return ret

    if "replacement" not in kwargs:
        replacement = True
    else:
        replacement = kwargs["replacement"]

    if isinstance(api_version, int):
        api_version = str(api_version)

    _function = (
        RANDOM_ORG_FUNCTIONS.get(api_version)
        .get("generateDecimalFractions")
        .get("method")
    )
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {
        "apiKey": api_key,
        "n": kwargs["number"],
        "decimalPlaces": kwargs["decimalPlaces"],
        "replacement": replacement,
    }

    result = _query(api_version=api_version, data=data)
    if result:
        if "random" in result:
            random_data = result.get("random").get("data")
            ret["data"] = random_data
        else:
            ret["res"] = False
            ret["message"] = result["message"]
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret


def generateGaussians(api_key=None, api_version=None, **kwargs):
    """
    This method generates true random numbers from a
    Gaussian distribution (also known as a normal distribution).

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :param number: How many random numbers you need.
                   Must be within the [1,1e4] range.
    :param mean: The distribution's mean. Must be
                 within the [-1e6,1e6] range.
    :param standardDeviation: The distribution's standard
                              deviation. Must be within
                              the [-1e6,1e6] range.
    :param significantDigits: The number of significant digits
                              to use. Must be within the [2,20] range.
    :return: The user list.

    CLI Example:

    .. code-block:: bash

        salt '*' random_org.generateGaussians number=10 mean=0.0 standardDeviation=1.0 significantDigits=8

    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    for item in ["number", "mean", "standardDeviation", "significantDigits"]:
        if item not in kwargs:
            ret["res"] = False
            ret["message"] = "Required argument, {} is missing.".format(item)
            return ret

    if not _numeric(kwargs["number"]) or not 1 <= kwargs["number"] <= 10000:
        ret["res"] = False
        ret["message"] = "Number of decimal fractions must be between 1 and 10000"
        return ret

    if not _numeric(kwargs["mean"]) or not -1000000 <= kwargs["mean"] <= 1000000:
        ret["res"] = False
        ret["message"] = "The distribution's mean must be between -1000000 and 1000000"
        return ret

    if (
        not _numeric(kwargs["standardDeviation"])
        or not -1000000 <= kwargs["standardDeviation"] <= 1000000
    ):
        ret["res"] = False
        ret[
            "message"
        ] = "The distribution's standard deviation must be between -1000000 and 1000000"
        return ret

    if (
        not _numeric(kwargs["significantDigits"])
        or not 2 <= kwargs["significantDigits"] <= 20
    ):
        ret["res"] = False
        ret["message"] = "The number of significant digits must be between 2 and 20"
        return ret

    if isinstance(api_version, int):
        api_version = str(api_version)

    _function = (
        RANDOM_ORG_FUNCTIONS.get(api_version).get("generateGaussians").get("method")
    )
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {
        "apiKey": api_key,
        "n": kwargs["number"],
        "mean": kwargs["mean"],
        "standardDeviation": kwargs["standardDeviation"],
        "significantDigits": kwargs["significantDigits"],
    }

    result = _query(api_version=api_version, data=data)
    if result:
        if "random" in result:
            random_data = result.get("random").get("data")
            ret["data"] = random_data
        else:
            ret["res"] = False
            ret["message"] = result["message"]
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret


def generateBlobs(api_key=None, api_version=None, **kwargs):
    """
    List all Slack users.

    :param api_key: The Random.org api key.
    :param api_version: The Random.org api version.
    :param format: Specifies the format in which the
                   blobs will be returned. Values
                   allowed are base64 and hex.
    :return: The user list.

    CLI Example:

    .. code-block:: bash

        salt '*' get_integers number=5 min=1 max=6

        salt '*' get_integers number=5 min=1 max=6
    """
    ret = {"res": True}

    if not api_key or not api_version:
        try:
            options = __salt__["config.option"]("random_org")
            if not api_key:
                api_key = options.get("api_key")
            if not api_version:
                api_version = options.get("api_version")
        except (NameError, KeyError, AttributeError):
            log.error("No Random.org api key found.")
            ret["message"] = "No Random.org api key or api version found."
            ret["res"] = False
            return ret

    for item in ["number", "size"]:
        if item not in kwargs:
            ret["res"] = False
            ret["message"] = "Required argument, {} is missing.".format(item)
            return ret

    if not _numeric(kwargs["number"]) or not 1 <= kwargs["number"] <= 100:
        ret["res"] = False
        ret["message"] = "Number of blobs must be between 1 and 100"
        return ret

    # size should be between range and divisible by 8
    if (
        not _numeric(kwargs["size"])
        or not 1 <= kwargs["size"] <= 1048576
        or kwargs["size"] % 8 != 0
    ):
        ret["res"] = False
        ret["message"] = "Number of blobs must be between 1 and 100"
        return ret

    if "format" in kwargs:
        _format = kwargs["format"]
        if _format not in ["base64", "hex"]:
            ret["res"] = False
            ret["message"] = "Format must be either base64 or hex."
            return ret
    else:
        _format = "base64"

    if isinstance(api_version, int):
        api_version = str(api_version)

    _function = RANDOM_ORG_FUNCTIONS.get(api_version).get("generateBlobs").get("method")
    data = {}
    data["id"] = 1911220
    data["jsonrpc"] = "2.0"
    data["method"] = _function
    data["params"] = {
        "apiKey": api_key,
        "n": kwargs["number"],
        "size": kwargs["size"],
        "format": _format,
    }

    result = _query(api_version=api_version, data=data)
    if result:
        if "random" in result:
            random_data = result.get("random").get("data")
            ret["data"] = random_data
        else:
            ret["res"] = False
            ret["message"] = result["message"]
    else:
        ret["res"] = False
        ret["message"] = result["message"]
    return ret



"""
tISM - the Immutable Secrets Manager SDB Module

:maintainer:    tISM
:maturity:      New
:platform:      all

.. versionadded:: 2017.7.0

This module will decrypt PGP encrypted secrets against a tISM server.

.. code::

  sdb://<profile>/<encrypted secret>

  sdb://tism/hQEMAzJ+GfdAB3KqAQf9E3cyvrPEWR1sf1tMvH0nrJ0bZa9kDFLPxvtwAOqlRiNp0F7IpiiVRF+h+sW5Mb4ffB1TElMzQ+/G5ptd6CjmgBfBsuGeajWmvLEi4lC6/9v1rYGjjLeOCCcN4Dl5AHlxUUaSrxB8akTDvSAnPvGhtRTZqDlltl5UEHsyYXM8RaeCrBw5Or1yvC9Ctx2saVp3xmALQvyhzkUv5pTb1mH0I9Z7E0ian07ZUOD+pVacDAf1oQcPpqkeNVTQQ15EP0fDuvnW+a0vxeLhkbFLfnwqhqEsvFxVFLHVLcs2ffE5cceeOMtVo7DS9fCtkdZr5hR7a+86n4hdKfwDMFXiBwSIPMkmY980N/H30L/r50+CBkuI/u4M2pXDcMYsvvt4ajCbJn91qaQ7BDI=

A profile must be setup in the minion configuration or pillar. If you want to
use sdb in a runner or pillar you must also place a profile in the master
configuration.

.. code-block:: yaml

    tism:
      driver: tism
      url: https://my.tismd:8080/decrypt
      token: eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJhZG1pbiI6MSwiZXhwIjoxNTg1MTExNDYwLCJqdGkiOiI3NnA5cWNiMWdtdmw4Iiwia2V5cyI6WyJBTEwiXX0.RtAhG6Uorf5xnSf4Ya_GwJnoHkCsql4r1_hiOeDSLzo
"""


import logging

import salt.utils.http as http
import salt.utils.json
from salt.exceptions import SaltConfigurationError

log = logging.getLogger(__name__)

__virtualname__ = "tism"


def __virtual__():
    """
    This module has no other system dependencies
    """
    return __virtualname__


def get(key, service=None, profile=None):  # pylint: disable=W0613
    """
    Get a decrypted secret from the tISMd API
    """

    if not profile.get("url") or not profile.get("token"):
        raise SaltConfigurationError(
            "url and/or token missing from the tism sdb profile"
        )

    request = {"token": profile["token"], "encsecret": key}

    result = http.query(
        profile["url"],
        method="POST",
        data=salt.utils.json.dumps(request),
    )

    decrypted = result.get("body")

    if not decrypted:
        log.warning(
            "tism.get sdb decryption request failed with error %s",
            result.get("error", "unknown"),
        )
        return "ERROR" + str(result.get("status", "unknown"))

    return decrypted



import pytest

import salt.utils.json
from salt.netapi.rest_tornado import saltnado


@pytest.fixture
def app_urls():
    return [
        (r"/jobs/(.*)", saltnado.JobsSaltAPIHandler),
        (r"/jobs", saltnado.JobsSaltAPIHandler),
    ]


@pytest.mark.slow_test
@pytest.mark.async_timeout(seconds=120)
async def test_get(http_client, subtests):
    # test with no JID
    response = await http_client.fetch("/jobs", method="GET", follow_redirects=False)
    response_obj = salt.utils.json.loads(response.body)["return"][0]
    assert response_obj
    assert isinstance(response_obj, dict)
    for ret in response_obj.values():
        with subtests.test('assert "Function" in ret'):
            assert "Function" in ret
        with subtests.test('assert "Target" in ret'):
            assert "Target" in ret
        with subtests.test('assert "Target-type" in ret'):
            assert "Target-type" in ret
        with subtests.test('assert "User" in ret'):
            assert "User" in ret
        with subtests.test('assert "StartTime" in ret'):
            assert "StartTime" in ret
        with subtests.test('assert "Arguments" in ret'):
            assert "Arguments" in ret

    # test with a specific JID passed in
    jid = next(iter(response_obj.keys()))
    response = await http_client.fetch(
        "/jobs/{}".format(jid),
        method="GET",
        follow_redirects=False,
    )
    response_obj = salt.utils.json.loads(response.body)["return"][0]
    assert response_obj
    assert isinstance(response_obj, dict)

    with subtests.test('assert "Function" in response_obj'):
        assert "Function" in response_obj
    with subtests.test('assert "Target" in response_obj'):
        assert "Target" in response_obj
    with subtests.test('assert "Target-type" in response_obj'):
        assert "Target-type" in response_obj
    with subtests.test('assert "User" in response_obj'):
        assert "User" in response_obj
    with subtests.test('assert "StartTime" in response_obj'):
        assert "StartTime" in response_obj
    with subtests.test('assert "Arguments" in response_obj'):
        assert "Arguments" in response_obj
    with subtests.test('assert "Result" in response_obj'):
        assert "Result" in response_obj



"""
Manage Dell DRAC from the Master

The login credentials need to be configured in the Salt master
configuration file.

.. code-block:: yaml

    drac:
      username: admin
      password: secret

"""


import logging

try:
    import paramiko

    HAS_PARAMIKO = True
except ImportError:
    HAS_PARAMIKO = False

log = logging.getLogger(__name__)


def __virtual__():
    if HAS_PARAMIKO:
        return True

    return (
        False,
        "The drac runner module cannot be loaded: paramiko package is not installed.",
    )


def __connect(hostname, timeout=20, username=None, password=None):
    """
    Connect to the DRAC
    """
    drac_cred = __opts__.get("drac")
    err_msg = (
        "No drac login credentials found. Please add the 'username' and 'password' "
        "fields beneath a 'drac' key in the master configuration file. Or you can "
        "pass in a username and password as kwargs at the CLI."
    )

    if not username:
        if drac_cred is None:
            log.error(err_msg)
            return False
        username = drac_cred.get("username", None)
    if not password:
        if drac_cred is None:
            log.error(err_msg)
            return False
        password = drac_cred.get("password", None)

    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        client.connect(hostname, username=username, password=password, timeout=timeout)
    except Exception as e:  # pylint: disable=broad-except
        log.error("Unable to connect to %s: %s", hostname, e)
        return False

    return client


def __version(client):
    """
    Grab DRAC version
    """
    versions = {
        9: "CMC",
        8: "iDRAC6",
        10: "iDRAC6",
        11: "iDRAC6",
        16: "iDRAC7",
        17: "iDRAC7",
    }

    if isinstance(client, paramiko.SSHClient):
        (stdin, stdout, stderr) = client.exec_command("racadm getconfig -g idRacInfo")

        for i in stdout.readlines():
            if i[2:].startswith("idRacType"):
                return versions.get(int(i[2:].split("=")[1]), None)

    return None


def pxe(hostname, timeout=20, username=None, password=None):
    """
    Connect to the Dell DRAC and have the boot order set to PXE
    and power cycle the system to PXE boot

    CLI Example:

    .. code-block:: bash

        salt-run drac.pxe example.com
    """
    _cmds = [
        "racadm config -g cfgServerInfo -o cfgServerFirstBootDevice pxe",
        "racadm config -g cfgServerInfo -o cfgServerBootOnce 1",
        "racadm serveraction powercycle",
    ]

    client = __connect(hostname, timeout, username, password)

    if isinstance(client, paramiko.SSHClient):
        for i, cmd in enumerate(_cmds, 1):
            log.info("Executing command %s", i)

            (stdin, stdout, stderr) = client.exec_command(cmd)

        if "successful" in stdout.readline():
            log.info("Executing command: %s", cmd)
        else:
            log.error("Unable to execute: %s", cmd)
            return False

    return True


def reboot(hostname, timeout=20, username=None, password=None):
    """
    Reboot a server using the Dell DRAC

    CLI Example:

    .. code-block:: bash

        salt-run drac.reboot example.com
    """
    client = __connect(hostname, timeout, username, password)

    if isinstance(client, paramiko.SSHClient):
        (stdin, stdout, stderr) = client.exec_command("racadm serveraction powercycle")

        if "successful" in stdout.readline():
            log.info("powercycle successful")
        else:
            log.error("powercycle racadm command failed")
            return False
    else:
        log.error("client was not of type paramiko.SSHClient")
        return False

    return True


def poweroff(hostname, timeout=20, username=None, password=None):
    """
    Power server off

    CLI Example:

    .. code-block:: bash

        salt-run drac.poweroff example.com
    """
    client = __connect(hostname, timeout, username, password)

    if isinstance(client, paramiko.SSHClient):
        (stdin, stdout, stderr) = client.exec_command("racadm serveraction powerdown")

        if "successful" in stdout.readline():
            log.info("powerdown successful")
        else:
            log.error("powerdown racadm command failed")
            return False
    else:
        log.error("client was not of type paramiko.SSHClient")
        return False

    return True


def poweron(hostname, timeout=20, username=None, password=None):
    """
    Power server on

    CLI Example:

    .. code-block:: bash

        salt-run drac.poweron example.com
    """
    client = __connect(hostname, timeout, username, password)

    if isinstance(client, paramiko.SSHClient):
        (stdin, stdout, stderr) = client.exec_command("racadm serveraction powerup")

        if "successful" in stdout.readline():
            log.info("powerup successful")
        else:
            log.error("powerup racadm command failed")
            return False
    else:
        log.error("client was not of type paramiko.SSHClient")
        return False

    return True


def version(hostname, timeout=20, username=None, password=None):
    """
    Display the version of DRAC

    CLI Example:

    .. code-block:: bash

        salt-run drac.version example.com
    """
    return __version(__connect(hostname, timeout, username, password))



"""
:codeauthor: Thayne Harbaugh (tharbaug@adobe.com)
"""
import logging
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time

import pytest
from pytestshellutils.utils.processes import ProcessResult, terminate_process

import salt.defaults.exitcodes
import salt.utils.path

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]


def test_context_retcode_salt(salt_cli, salt_minion):
    """
    Test that a nonzero retcode set in the context dunder will cause the
    salt CLI to set a nonzero retcode.
    """
    # test.retcode will set the retcode in the context dunder
    ret = salt_cli.run("test.retcode", "0", minion_tgt=salt_minion.id)
    assert ret.returncode == 0, ret
    ret = salt_cli.run("test.retcode", "42", minion_tgt=salt_minion.id)
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret


def test_salt_error(salt_cli, salt_minion):
    """
    Test that we return the expected retcode when a minion function raises
    an exception.
    """
    ret = salt_cli.run("test.raise_exception", "TypeError", minion_tgt=salt_minion.id)
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_cli.run(
        "test.raise_exception",
        "salt.exceptions.CommandNotFoundError",
        minion_tgt=salt_minion.id,
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_cli.run(
        "test.raise_exception",
        "salt.exceptions.CommandExecutionError",
        minion_tgt=salt_minion.id,
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_cli.run(
        "test.raise_exception",
        "salt.exceptions.SaltInvocationError",
        minion_tgt=salt_minion.id,
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_cli.run(
        "test.raise_exception",
        "OSError",
        "2",
        '"No such file or directory" /tmp/foo.txt',
        minion_tgt=salt_minion.id,
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_cli.run(
        "test.echo", "{foo: bar, result: False}", minion_tgt=salt_minion.id
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_cli.run(
        "test.echo", "{foo: bar, success: False}", minion_tgt=salt_minion.id
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret


def test_missing_minion(salt_cli, salt_master, salt_minion):
    """
    Test that a minion which doesn't respond results in a nonzeo exit code
    """
    good = salt.utils.path.join(
        salt_master.config["pki_dir"], "minions", salt_minion.id
    )
    bad = salt.utils.path.join(salt_master.config["pki_dir"], "minions", "minion2")
    try:
        # Copy the key
        shutil.copyfile(good, bad)
        ret = salt_cli.run(
            "--timeout=5", "test.ping", minion_tgt="minion2", _timeout=120
        )
        assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret
    finally:
        # Now get rid of it
        try:
            os.remove(bad)
        except OSError as exc:
            if exc.errno != os.errno.ENOENT:
                log.error(
                    "Failed to remove %s, this may affect other tests: %s", bad, exc
                )


def test_exit_status_unknown_argument(salt_cli):
    """
    Ensure correct exit status when an unknown argument is passed to salt CLI.
    """
    ret = salt_cli.run(
        "--unknown-argument", minion_tgt="minion-tgt-is-mandatory-by-salt-factories"
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_USAGE, ret
    assert "Usage" in ret.stderr
    assert "no such option: --unknown-argument" in ret.stderr


def test_exit_status_correct_usage(salt_cli, salt_minion):
    """
    Ensure correct exit status when salt CLI starts correctly.

    """
    ret = salt_cli.run("test.ping", minion_tgt=salt_minion.id)
    assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


@pytest.mark.skip_on_windows(reason="Windows does not support SIGINT")
@pytest.mark.skip_initial_onedir_failure
def test_interrupt_on_long_running_job(salt_cli, salt_master, salt_minion):
    """
    Ensure that a call to ``salt`` that is taking too long, when a user
    hits CTRL-C, that the JID is printed to the console.

    Refer to https://github.com/saltstack/salt/issues/60963 for more details
    """
    # Ensure test.sleep is working as supposed
    start = time.time()
    ret = salt_cli.run("test.sleep", "1", minion_tgt=salt_minion.id)
    stop = time.time()
    assert ret.returncode == 0
    assert ret.data is True
    assert stop - start > 1, "The command should have taken more than 1 second"

    # Now the real test
    terminal_stdout = tempfile.SpooledTemporaryFile(512000, buffering=0)
    terminal_stderr = tempfile.SpooledTemporaryFile(512000, buffering=0)
    cmdline = [
        sys.executable,
        salt_cli.get_script_path(),
        "--config-dir={}".format(salt_master.config_dir),
        salt_minion.id,
        "test.sleep",
        "30",
    ]

    # If this test starts failing, commend the following block of code
    proc = subprocess.Popen(
        cmdline,
        shell=False,
        stdout=terminal_stdout,
        stderr=terminal_stderr,
        universal_newlines=True,
    )
    # and uncomment the following block of code

    # with default_signals(signal.SIGINT, signal.SIGTERM):
    #    proc = subprocess.Popen(
    #        cmdline,
    #        shell=False,
    #        stdout=terminal_stdout,
    #        stderr=terminal_stderr,
    #        universal_newlines=True,
    #    )

    # What this means is that something in salt or the test suite is setting
    # the SIGTERM and SIGINT signals to SIG_IGN, ignore.
    # Check which line of code is doing that and fix it
    start = time.time()
    try:
        # Make sure it actually starts
        proc.wait(1)
    except subprocess.TimeoutExpired:
        pass
    else:
        terminate_process(proc.pid, kill_children=True)
        pytest.fail("The test process failed to start")

    time.sleep(2)
    # Send CTRL-C to the process
    os.kill(proc.pid, signal.SIGINT)
    with proc:
        # Wait for the process to terminate, to avoid zombies.
        # Shouldn't really take the 30 seconds
        proc.wait(30)
        # poll the terminal so the right returncode is set on the popen object
        proc.poll()
        # This call shouldn't really be necessary
        proc.communicate()
    stop = time.time()

    terminal_stdout.flush()
    terminal_stdout.seek(0)
    stdout = proc._translate_newlines(
        terminal_stdout.read(), __salt_system_encoding__, sys.stdout.errors
    )
    terminal_stdout.close()

    terminal_stderr.flush()
    terminal_stderr.seek(0)
    stderr = proc._translate_newlines(
        terminal_stderr.read(), __salt_system_encoding__, sys.stderr.errors
    )
    terminal_stderr.close()
    ret = ProcessResult(
        returncode=proc.returncode, stdout=stdout, stderr=stderr, cmdline=proc.args
    )
    log.debug(ret)
    # If the minion ID is on stdout it means that the command finished and wasn't terminated
    assert (
        salt_minion.id not in ret.stdout
    ), "The command wasn't actually terminated. Took {} seconds.".format(
        round(stop - start, 2)
    )

    # Make sure the ctrl+c exited gracefully
    assert "Exiting gracefully on Ctrl-c" in ret.stderr
    assert "Exception ignored in" not in ret.stderr
    assert "This job's jid is" in ret.stderr



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import logging
import os

import salt.modules.dpkg_lowpkg as dpkg
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

DPKG_ERROR_MSG = """dpkg-query: package 'httpd' is not installed
Use dpkg --contents (= dpkg-deb --contents) to list archive files contents.
"""

DPKG_L_OUTPUT = {
    "hostname": """\
/.
/bin
/bin/hostname
/usr
/usr/share
/usr/share/doc
/usr/share/doc/hostname
/usr/share/doc/hostname/changelog.gz
/usr/share/doc/hostname/copyright
/usr/share/man
/usr/share/man/man1
/usr/share/man/man1/hostname.1.gz
/bin/dnsdomainname
/bin/domainname
/bin/nisdomainname
/bin/ypdomainname
/usr/share/man/man1/dnsdomainname.1.gz
/usr/share/man/man1/domainname.1.gz
/usr/share/man/man1/nisdomainname.1.gz
/usr/share/man/man1/ypdomainname.1.gz
"""
}


class DpkgTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.dpkg
    """

    def setUp(self):
        dpkg_lowpkg_logger = logging.getLogger("salt.modules.dpkg_lowpkg")
        self.level = dpkg_lowpkg_logger.level
        dpkg_lowpkg_logger.setLevel(logging.FATAL)

    def tearDown(self):
        logging.getLogger("salt.modules.dpkg_lowpkg").setLevel(self.level)

    def dpkg_L_side_effect(self, cmd, **kwargs):
        self.assertEqual(cmd[:2], ["dpkg", "-L"])
        package = cmd[2]
        return DPKG_L_OUTPUT[package]

    def setup_loader_modules(self):
        return {dpkg: {}}

    # 'unpurge' function tests: 2

    def test_unpurge(self):
        """
        Test if it change package selection for each package
        specified to 'install'
        """
        mock = MagicMock(return_value=[])
        with patch.dict(dpkg.__salt__, {"pkg.list_pkgs": mock, "cmd.run": mock}):
            self.assertDictEqual(dpkg.unpurge("curl"), {})

    def test_unpurge_empty_package(self):
        """
        Test if it change package selection for each package
        specified to 'install'
        """
        self.assertDictEqual(dpkg.unpurge(), {})

    # 'list_pkgs' function tests: 1

    def test_list_pkgs(self):
        """
        Test if it lists the packages currently installed
        """
        mock = MagicMock(
            return_value={
                "retcode": 0,
                "stderr": "",
                "stdout": "installed\thostname\t3.21",
            }
        )
        with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(dpkg.list_pkgs("hostname"), {"hostname": "3.21"})

        mock = MagicMock(
            return_value={
                "retcode": 1,
                "stderr": "dpkg-query: no packages found matching httpd",
                "stdout": "",
            }
        )
        with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(
                dpkg.list_pkgs("httpd"),
                "Error:  dpkg-query: no packages found matching httpd",
            )

    # 'file_list' function tests: 1

    def test_file_list(self):
        """
        Test if it lists the files that belong to a package.
        """
        dpkg_query_mock = MagicMock(
            return_value={"retcode": 0, "stderr": "", "stdout": "installed\thostname"}
        )
        dpkg_L_mock = MagicMock(side_effect=self.dpkg_L_side_effect)
        with patch.dict(
            dpkg.__salt__, {"cmd.run_all": dpkg_query_mock, "cmd.run": dpkg_L_mock}
        ):
            self.assertDictEqual(
                dpkg.file_list("hostname"),
                {
                    "errors": [],
                    "files": [
                        "/.",
                        "/bin",
                        "/bin/dnsdomainname",
                        "/bin/domainname",
                        "/bin/hostname",
                        "/bin/nisdomainname",
                        "/bin/ypdomainname",
                        "/usr",
                        "/usr/share",
                        "/usr/share/doc",
                        "/usr/share/doc/hostname",
                        "/usr/share/doc/hostname/changelog.gz",
                        "/usr/share/doc/hostname/copyright",
                        "/usr/share/man",
                        "/usr/share/man/man1",
                        "/usr/share/man/man1/dnsdomainname.1.gz",
                        "/usr/share/man/man1/domainname.1.gz",
                        "/usr/share/man/man1/hostname.1.gz",
                        "/usr/share/man/man1/nisdomainname.1.gz",
                        "/usr/share/man/man1/ypdomainname.1.gz",
                    ],
                },
            )

        mock = MagicMock(
            return_value={"retcode": 1, "stderr": DPKG_ERROR_MSG, "stdout": ""}
        )
        with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(dpkg.file_list("httpd"), "Error:  " + DPKG_ERROR_MSG)

    # 'file_dict' function tests: 1

    def test_file_dict(self):
        """
        Test if it lists the files that belong to a package, grouped by package
        """
        dpkg_query_mock = MagicMock(
            return_value={"retcode": 0, "stderr": "", "stdout": "installed\thostname"}
        )
        dpkg_L_mock = MagicMock(side_effect=self.dpkg_L_side_effect)
        with patch.dict(
            dpkg.__salt__, {"cmd.run_all": dpkg_query_mock, "cmd.run": dpkg_L_mock}
        ):
            expected = {
                "errors": [],
                "packages": {
                    "hostname": [
                        "/.",
                        "/bin",
                        "/bin/hostname",
                        "/usr",
                        "/usr/share",
                        "/usr/share/doc",
                        "/usr/share/doc/hostname",
                        "/usr/share/doc/hostname/changelog.gz",
                        "/usr/share/doc/hostname/copyright",
                        "/usr/share/man",
                        "/usr/share/man/man1",
                        "/usr/share/man/man1/hostname.1.gz",
                        "/bin/dnsdomainname",
                        "/bin/domainname",
                        "/bin/nisdomainname",
                        "/bin/ypdomainname",
                        "/usr/share/man/man1/dnsdomainname.1.gz",
                        "/usr/share/man/man1/domainname.1.gz",
                        "/usr/share/man/man1/nisdomainname.1.gz",
                        "/usr/share/man/man1/ypdomainname.1.gz",
                    ]
                },
            }
            self.assertDictEqual(dpkg.file_dict("hostname"), expected)

        mock = MagicMock(
            return_value={"retcode": 1, "stderr": DPKG_ERROR_MSG, "stdout": ""}
        )
        with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(dpkg.file_dict("httpd"), "Error:  " + DPKG_ERROR_MSG)

    def test_bin_pkg_info_spaces(self):
        """
        Test the bin_pkg_info function
        """
        file_proto_mock = MagicMock(return_value=True)
        with patch.dict(dpkg.__salt__, {"config.valid_fileproto": file_proto_mock}):
            cache_mock = MagicMock(return_value="/path/to/some/package.deb")
            with patch.dict(dpkg.__salt__, {"cp.cache_file": cache_mock}):
                dpkg_info_mock = MagicMock(
                    return_value={
                        "retcode": 0,
                        "stderr": "",
                        "stdout": (
                            " new Debian package, version 2.0\n"
                            " size 123456 bytes: control archive: 4029  bytes.\n"
                            " Package          : package_name\n"
                            " Version          : 1.0\n"
                            " Section          : section_name\n"
                            " Priority         : priority\n"
                            " Architecture     : all\n"
                            " Description      : some package\n"
                        ),
                    }
                )
                with patch.dict(dpkg.__salt__, {"cmd.run_all": dpkg_info_mock}):
                    self.assertEqual(
                        dpkg.bin_pkg_info("package.deb")["name"], "package_name"
                    )

    def test_bin_pkg_info_no_spaces(self):
        """
        Test the bin_pkg_info function
        """
        file_proto_mock = MagicMock(return_value=True)
        with patch.dict(dpkg.__salt__, {"config.valid_fileproto": file_proto_mock}):
            cache_mock = MagicMock(return_value="/path/to/some/package.deb")
            with patch.dict(dpkg.__salt__, {"cp.cache_file": cache_mock}):
                dpkg_info_mock = MagicMock(
                    return_value={
                        "retcode": 0,
                        "stderr": "",
                        "stdout": (
                            " new Debian package, version 2.0\n"
                            " size 123456 bytes: control archive: 4029  bytes.\n"
                            " Package: package_name\n"
                            " Version: 1.0\n"
                            " Section: section_name\n"
                            " Priority: priority\n"
                            " Architecture: all\n"
                            " Description: some package\n"
                        ),
                    }
                )
                with patch.dict(dpkg.__salt__, {"cmd.run_all": dpkg_info_mock}):
                    self.assertEqual(
                        dpkg.bin_pkg_info("package.deb")["name"], "package_name"
                    )

    def test_info(self):
        """
        Test package info
        """
        mock = MagicMock(
            return_value={
                "retcode": 0,
                "stderr": "",
                "stdout": os.linesep.join(
                    [
                        "package:bash",
                        "revision:",
                        "architecture:amd64",
                        "maintainer:Ubuntu Developers"
                        " <ubuntu-devel-discuss@lists.ubuntu.com>",
                        "summary:",
                        "source:bash",
                        "version:4.4.18-2ubuntu1",
                        "section:shells",
                        "installed_size:1588",
                        "size:",
                        "MD5:",
                        "SHA1:",
                        "SHA256:",
                        "origin:",
                        "homepage:http://tiswww.case.edu/php/chet/bash/bashtop.html",
                        "status:ii ",
                        "description:GNU Bourne Again SHell",
                        " Bash is an sh-compatible command language interpreter that"
                        " executes",
                        " commands read from the standard input or from a file.  Bash"
                        " also",
                        " incorporates useful features from the Korn and C shells (ksh"
                        " and csh).",
                        " .",
                        " Bash is ultimately intended to be a conformant implementation"
                        " of the",
                        " IEEE POSIX Shell and Tools specification (IEEE Working Group"
                        " 1003.2).",
                        " .",
                        " The Programmable Completion Code, by Ian Macdonald, is now"
                        " found in",
                        " the bash-completion package.",
                        "",
                        "*/~^\\*",  # pylint: disable=W1401
                    ]
                ),
            }
        )

        with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}), patch.dict(
            dpkg.__grains__, {"os": "Ubuntu", "osrelease_info": (18, 4)}
        ), patch("salt.utils.path.which", MagicMock(return_value=False)), patch(
            "os.path.exists", MagicMock(return_value=False)
        ), patch(
            "os.path.getmtime", MagicMock(return_value=1560199259.0)
        ):
            self.assertDictEqual(
                dpkg.info("bash"),
                {
                    "bash": {
                        "architecture": "amd64",
                        "description": os.linesep.join(
                            [
                                "GNU Bourne Again SHell",
                                " Bash is an sh-compatible command language interpreter"
                                " that executes",
                                " commands read from the standard input or from a file."
                                "  Bash also",
                                " incorporates useful features from the Korn and C"
                                " shells (ksh and csh).",
                                " .",
                                " Bash is ultimately intended to be a conformant"
                                " implementation of the",
                                " IEEE POSIX Shell and Tools specification (IEEE"
                                " Working Group 1003.2).",
                                " .",
                                " The Programmable Completion Code, by Ian Macdonald,"
                                " is now found in",
                                " the bash-completion package." + os.linesep,
                            ]
                        ),
                        "homepage": "http://tiswww.case.edu/php/chet/bash/bashtop.html",
                        "maintainer": (
                            "Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>"
                        ),
                        "package": "bash",
                        "section": "shells",
                        "source": "bash",
                        "status": "ii",
                        "version": "4.4.18-2ubuntu1",
                    }
                },
            )



"""
Tests for the Openstack Cloud Provider
"""


import logging
import os
import shutil
from time import sleep

import pytest
from saltfactories.utils import random_string

import salt.utils.files
from salt.config import cloud_config, cloud_providers_config
from salt.utils.yaml import safe_load
from tests.support.case import ShellCase
from tests.support.paths import FILES
from tests.support.runtests import RUNTIME_VARS

TIMEOUT = 500

log = logging.getLogger(__name__)


@pytest.mark.expensive_test
class CloudTest(ShellCase):
    PROVIDER = ""
    REQUIRED_PROVIDER_CONFIG_ITEMS = tuple()
    __RE_RUN_DELAY = 30
    __RE_TRIES = 12

    @staticmethod
    def clean_cloud_dir(tmp_dir):
        """
        Clean the cloud.providers.d tmp directory
        """
        # make sure old provider configs are deleted
        if not os.path.isdir(tmp_dir):
            return
        for fname in os.listdir(tmp_dir):
            os.remove(os.path.join(tmp_dir, fname))

    def query_instances(self):
        """
        Standardize the data returned from a salt-cloud --query
        """
        return {
            x.strip(": ")
            for x in self.run_cloud("--query")
            if x.lstrip().lower().startswith("cloud-test-")
        }

    def _instance_exists(self, instance_name=None, query=None):
        """
        :param instance_name: The name of the instance to check for in salt-cloud.
        For example this is may used when a test temporarily renames an instance
        :param query: The result of a salt-cloud --query run outside of this function
        """
        if not instance_name:
            instance_name = self.instance_name
        if not query:
            query = self.query_instances()

        log.debug('Checking for "%s" in %s', instance_name, query)
        if isinstance(query, set):
            return instance_name in query
        return any(instance_name == q.strip(": ") for q in query)

    def assertInstanceExists(self, creation_ret=None, instance_name=None):
        """
        :param instance_name: Override the checked instance name, otherwise the class default will be used.
        :param creation_ret: The return value from the run_cloud() function that created the instance
        """
        if not instance_name:
            instance_name = self.instance_name

        # If it exists but doesn't show up in the creation_ret, there was probably an error during creation
        if creation_ret:
            self.assertIn(
                instance_name,
                [i.strip(": ") for i in creation_ret],
                "An error occured during instance creation:  |\n\t{}\n\t|".format(
                    "\n\t".join(creation_ret)
                ),
            )
        else:
            # Verify that the instance exists via query
            query = self.query_instances()
            for tries in range(self.__RE_TRIES):
                if self._instance_exists(instance_name, query):
                    log.debug(
                        'Instance "%s" reported after %s seconds',
                        instance_name,
                        tries * self.__RE_RUN_DELAY,
                    )
                    break
                else:
                    sleep(self.__RE_RUN_DELAY)
                    query = self.query_instances()

            # Assert that the last query was successful
            self.assertTrue(
                self._instance_exists(instance_name, query),
                'Instance "{}" was not created successfully: {}'.format(
                    self.instance_name, ", ".join(query)
                ),
            )

            log.debug('Instance exists and was created: "%s"', instance_name)

    def assertDestroyInstance(self, instance_name=None, timeout=None):
        if timeout is None:
            timeout = TIMEOUT
        if not instance_name:
            instance_name = self.instance_name
        log.debug('Deleting instance "%s"', instance_name)
        delete_str = self.run_cloud(
            "-d {} --assume-yes --out=yaml".format(instance_name), timeout=timeout
        )
        if delete_str:
            delete = safe_load("\n".join(delete_str))
            self.assertIn(self.profile_str, delete)
            self.assertIn(self.PROVIDER, delete[self.profile_str])
            self.assertIn(instance_name, delete[self.profile_str][self.PROVIDER])

            delete_status = delete[self.profile_str][self.PROVIDER][instance_name]
            if isinstance(delete_status, str):
                self.assertEqual(delete_status, "True")
                return
            elif isinstance(delete_status, dict):
                current_state = delete_status.get("currentState")
                if current_state:
                    if current_state.get("ACTION"):
                        self.assertIn(".delete", current_state.get("ACTION"))
                        return
                    else:
                        self.assertEqual(current_state.get("name"), "shutting-down")
                        return
        # It's not clear from the delete string that deletion was successful, ask salt-cloud after a delay
        query = self.query_instances()
        # some instances take a while to report their destruction
        for tries in range(6):
            if self._instance_exists(query=query):
                sleep(30)
                log.debug(
                    'Instance "%s" still found in query after %s tries: %s',
                    instance_name,
                    tries,
                    query,
                )
                query = self.query_instances()
        # The last query should have been successful
        self.assertNotIn(instance_name, self.query_instances())

    @property
    def instance_name(self):
        if not hasattr(self, "_instance_name"):
            # Create the cloud instance name to be used throughout the tests
            subclass = self.__class__.__name__.strip("Test")
            # Use the first three letters of the subclass, fill with '-' if too short
            self._instance_name = random_string(
                "cloud-test-{:-<3}-".format(subclass[:3]), uppercase=False
            ).lower()
        return self._instance_name

    @property
    def providers(self):
        if not hasattr(self, "_providers"):
            self._providers = self.run_cloud("--list-providers")
        return self._providers

    @property
    def provider_config(self):
        if not hasattr(self, "_provider_config"):
            self._provider_config = cloud_providers_config(
                os.path.join(
                    RUNTIME_VARS.TMP_CONF_DIR,
                    "cloud.providers.d",
                    self.PROVIDER + ".conf",
                )
            )
        return self._provider_config[self.profile_str][self.PROVIDER]

    @property
    def config(self):
        if not hasattr(self, "_config"):
            self._config = cloud_config(
                os.path.join(
                    RUNTIME_VARS.TMP_CONF_DIR,
                    "cloud.profiles.d",
                    self.PROVIDER + ".conf",
                )
            )
        return self._config

    @property
    def profile_str(self):
        return self.PROVIDER + "-config"

    def add_profile_config(self, name, data, conf, new_profile):
        """
        copy the current profile and add a new profile in the same file
        """
        conf_path = os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "cloud.profiles.d", conf)
        with salt.utils.files.fopen(conf_path, "r") as fp:
            conf = safe_load(fp)
        conf[new_profile] = conf[name].copy()
        conf[new_profile].update(data)
        with salt.utils.files.fopen(conf_path, "w") as fp:
            salt.utils.yaml.safe_dump(conf, fp)

    def setUp(self):
        """
        Sets up the test requirements.  In child classes, define PROVIDER and REQUIRED_PROVIDER_CONFIG_ITEMS or this will fail
        """
        super().setUp()

        if not self.PROVIDER:
            self.fail("A PROVIDER must be defined for this test")

        # check if appropriate cloud provider and profile files are present
        if self.profile_str + ":" not in self.providers:
            self.skipTest(
                "Configuration file for {0} was not found. Check {0}.conf files "
                "in tests/integration/files/conf/cloud.*.d/ to run these tests.".format(
                    self.PROVIDER
                )
            )

        missing_conf_item = []
        for att in self.REQUIRED_PROVIDER_CONFIG_ITEMS:
            if not self.provider_config.get(att):
                missing_conf_item.append(att)

        if missing_conf_item:
            self.skipTest(
                "Conf items are missing that must be provided to run these tests:  {}".format(
                    ", ".join(missing_conf_item)
                )
                + "\nCheck tests/integration/files/conf/cloud.providers.d/{}.conf".format(
                    self.PROVIDER
                )
            )

    def _alt_names(self):
        """
        Check for an instances created alongside this test's instance that weren't cleaned up
        """
        query = self.query_instances()
        instances = set()
        for q in query:
            # Verify but this is a new name and not a shutting down ec2 instance
            if q.startswith(self.instance_name) and not q.split("-")[-1].startswith(
                "DEL"
            ):
                instances.add(q)
                log.debug(
                    'Adding "%s" to the set of instances that needs to be deleted', q
                )
        return instances

    def _ensure_deletion(self, instance_name=None):
        """
        Make sure that the instance absolutely gets deleted, but fail the test if it happens in the tearDown
        :return True if an instance was deleted, False if no instance was deleted; and a message
        """
        destroyed = False
        if not instance_name:
            instance_name = self.instance_name

        if self._instance_exists(instance_name):
            for tries in range(3):
                try:
                    self.assertDestroyInstance(instance_name)
                    return (
                        False,
                        'The instance "{}" was deleted during the tearDown, not the'
                        " test.".format(instance_name),
                    )
                except AssertionError as e:
                    log.error(
                        'Failed to delete instance "%s". Tries: %s\n%s',
                        instance_name,
                        tries,
                        str(e),
                    )
                if not self._instance_exists():
                    destroyed = True
                    break
                else:
                    sleep(30)

            if not destroyed:
                # Destroying instances in the tearDown is a contingency, not the way things should work by default.
                return (
                    False,
                    'The Instance "{}" was not deleted after multiple attempts'.format(
                        instance_name
                    ),
                )

        return (
            True,
            'The instance "{}" cleaned up properly after the test'.format(
                instance_name
            ),
        )

    def tearDown(self):
        """
        Clean up after tests, If the instance still exists for any reason, delete it.
        Instances should be destroyed before the tearDown, assertDestroyInstance() should be called exactly
        one time in a test for each instance created.  This is a failSafe and something went wrong
        if the tearDown is where an instance is destroyed.
        """
        success = True
        fail_messages = []
        alt_names = self._alt_names()
        for instance in alt_names:
            alt_destroyed, alt_destroy_message = self._ensure_deletion(instance)
            if not alt_destroyed:
                success = False
                fail_messages.append(alt_destroy_message)
                log.error(
                    'Failed to destroy instance "%s": %s', instance, alt_destroy_message
                )
        self.assertTrue(success, "\n".join(fail_messages))
        self.assertFalse(
            alt_names, "Cleanup should happen in the test, not the TearDown"
        )

    @classmethod
    def tearDownClass(cls):
        cls.clean_cloud_dir(cls.tmp_provider_dir)

    @classmethod
    def setUpClass(cls):
        # clean up before setup
        cls.tmp_provider_dir = os.path.join(
            RUNTIME_VARS.TMP_CONF_DIR, "cloud.providers.d"
        )
        cls.clean_cloud_dir(cls.tmp_provider_dir)

        # add the provider config for only the cloud we are testing
        provider_file = cls.PROVIDER + ".conf"
        shutil.copyfile(
            os.path.join(
                os.path.join(FILES, "conf", "cloud.providers.d"), provider_file
            ),
            os.path.join(os.path.join(cls.tmp_provider_dir, provider_file)),
        )



"""
    tests.pytests.integration.conftest
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    PyTest fixtures
"""

import logging

import pytest

log = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def salt_master(salt_master_factory):
    """
    A running salt-master fixture
    """
    with salt_master_factory.started():
        yield salt_master_factory


@pytest.fixture(scope="package")
def salt_minion(salt_master, salt_minion_factory):
    """
    A running salt-minion fixture
    """
    assert salt_master.is_running()
    with salt_minion_factory.started():
        # Sync All
        salt_call_cli = salt_minion_factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield salt_minion_factory


@pytest.fixture(scope="module")
def salt_sub_minion(salt_master, salt_sub_minion_factory):
    """
    A second running salt-minion fixture
    """
    assert salt_master.is_running()
    with salt_sub_minion_factory.started():
        # Sync All
        salt_call_cli = salt_sub_minion_factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield salt_sub_minion_factory


@pytest.fixture(scope="package")
def salt_cli(salt_master):
    """
    The ``salt`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_cli(timeout=30)


@pytest.fixture(scope="package")
def salt_call_cli(salt_minion):
    """
    The ``salt-call`` CLI as a fixture against the running minion
    """
    assert salt_minion.is_running()
    return salt_minion.salt_call_cli(timeout=30)


@pytest.fixture(scope="package")
def salt_cp_cli(salt_master):
    """
    The ``salt-cp`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_cp_cli(timeout=30)


@pytest.fixture(scope="package")
def salt_key_cli(salt_master):
    """
    The ``salt-key`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_key_cli(timeout=30)


@pytest.fixture(scope="package")
def salt_run_cli(salt_master):
    """
    The ``salt-run`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_run_cli(timeout=30)


@pytest.fixture(scope="module")
def salt_ssh_cli(salt_master, salt_ssh_roster_file, sshd_config_dir):
    """
    The ``salt-ssh`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_ssh_cli(
        timeout=180,
        roster_file=salt_ssh_roster_file,
        target_host="localhost",
        client_key=str(sshd_config_dir / "client_key"),
        base_script_args=["--ignore-host-keys"],
    )



"""
    :codeauthor: :email:`Gareth J. Greenaway <ggreenaway@vmware.com>`
"""

import logging
import os

import pytest

import salt.modules.beacons as beacons
from salt.utils.event import SaltEvent
from tests.support.mock import MagicMock, call, mock_open, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {beacons: {"__opts__": minion_opts}}


@pytest.mark.slow_test
def test_delete():
    """
    Test deleting a beacon.
    """
    comm1 = "Deleted beacon: ps."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_delete_complete",
            "beacons": {},
        },
    ]

    with patch.dict(
        beacons.__opts__,
        {
            "beacons": {
                "ps": [{"processes": {"salt-master": "stopped", "apache2": "stopped"}}]
            },
        },
    ):
        mock = MagicMock(return_value=True)
        with patch.dict(beacons.__salt__, {"event.fire": mock}):
            with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                assert beacons.delete("ps") == {"comment": comm1, "result": True}


@pytest.mark.slow_test
def test_add():
    """
    Test adding a beacon
    """
    comm1 = "Added beacon: ps."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_complete",
            "beacons": {},
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_available_complete",
            "beacons": ["ps"],
        },
        {
            "complete": True,
            "valid": True,
            "vcomment": "",
            "tag": "/salt/minion/minion_beacons_list_complete",
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacon_add_complete",
            "beacons": {
                "ps": [{"processes": {"salt-master": "stopped", "apache2": "stopped"}}]
            },
        },
    ]

    mock = MagicMock(return_value=True)
    with patch.dict(beacons.__salt__, {"event.fire": mock}):
        with patch.object(SaltEvent, "get_event", side_effect=event_returns):
            assert beacons.add(
                "ps",
                [
                    {
                        "processes": {
                            "salt-master": "stopped",
                            "apache2": "stopped",
                        }
                    }
                ],
            ) == {"comment": comm1, "result": True}


@pytest.mark.slow_test
def test_save():
    """
    Test saving beacons.
    """
    _beacon_conf_file = os.path.join(
        os.path.dirname(beacons.__opts__["conf_file"]),
        os.path.dirname(beacons.__opts__["default_include"]),
        "beacons.conf",
    )
    _beacons_data = {
        "ps": [{"processes": {"salt-master": "stopped", "apache2": "stopped"}}]
    }

    # Test that beacons contents are written to config file.
    _expected = {
        "comment": "Beacons saved to {}.".format(_beacon_conf_file),
        "result": True,
    }
    with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
        with patch.object(beacons, "list_", MagicMock(return_value=_beacons_data)):
            ret = beacons.save()
            assert ret == _expected

            _call = call(
                "beacons:\n  ps:\n  - processes:\n      apache2: stopped\n      salt-master: stopped\n"
            )
            write_calls = fopen_mock.filehandles[_beacon_conf_file][
                0
            ].write._mock_mock_calls
            assert _call in write_calls

    _beacons_data = {}

    # Test that when beacons is empty then an empty config file is written.
    _expected = {
        "comment": "Beacons saved to {}.".format(_beacon_conf_file),
        "result": True,
    }
    with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
        with patch.object(beacons, "list_", MagicMock(return_value=_beacons_data)):
            ret = beacons.save()
            assert ret == _expected

            _call = call("")
            write_calls = fopen_mock.filehandles[_beacon_conf_file][
                0
            ].write._mock_mock_calls
            assert _call in write_calls


@pytest.mark.slow_test
def test_disable():
    """
    Test disabling beacons
    """
    comm1 = "Disabled beacons on minion."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_disabled_complete",
            "beacons": {
                "enabled": False,
                "ps": [{"processes": {"salt-master": "stopped", "apache2": "stopped"}}],
            },
        }
    ]

    mock = MagicMock(return_value=True)
    with patch.dict(beacons.__salt__, {"event.fire": mock}):
        with patch.object(SaltEvent, "get_event", side_effect=event_returns):
            assert beacons.disable() == {"comment": comm1, "result": True}


@pytest.mark.slow_test
def test_enable():
    """
    Test enabling beacons
    """
    comm1 = "Enabled beacons on minion."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacon_enabled_complete",
            "beacons": {
                "enabled": True,
                "ps": [{"processes": {"salt-master": "stopped", "apache2": "stopped"}}],
            },
        }
    ]

    mock = MagicMock(return_value=True)
    with patch.dict(beacons.__salt__, {"event.fire": mock}):
        with patch.object(SaltEvent, "get_event", side_effect=event_returns):
            assert beacons.enable() == {"comment": comm1, "result": True}


@pytest.mark.slow_test
def test_add_beacon_module():
    """
    Test adding a beacon
    """
    comm1 = "Added beacon: watch_salt_master."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_complete",
            "beacons": {},
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_available_complete",
            "beacons": ["ps"],
        },
        {
            "complete": True,
            "valid": True,
            "vcomment": "",
            "tag": "/salt/minion/minion_beacons_list_complete",
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacon_add_complete",
            "beacons": {
                "watch_salt_master": [
                    {"processes": {"salt-master": "stopped"}},
                    {"beacon_module": "ps"},
                ]
            },
        },
    ]

    mock = MagicMock(return_value=True)
    with patch.dict(beacons.__salt__, {"event.fire": mock}):
        with patch.object(SaltEvent, "get_event", side_effect=event_returns):
            assert beacons.add(
                "watch_salt_master",
                [
                    {"processes": {"salt-master": "stopped"}},
                    {"beacon_module": "ps"},
                ],
            ) == {"comment": comm1, "result": True}


@pytest.mark.slow_test
def test_enable_beacon_module():
    """
    Test enabling beacons
    """
    comm1 = "Enabled beacons on minion."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacon_enabled_complete",
            "beacons": {
                "enabled": True,
                "watch_salt_master": [
                    {"processes": {"salt-master": "stopped"}},
                    {"beacon_module": "ps"},
                ],
            },
        }
    ]

    mock = MagicMock(return_value=True)
    with patch.dict(beacons.__salt__, {"event.fire": mock}):
        with patch.object(SaltEvent, "get_event", side_effect=event_returns):
            assert beacons.enable() == {"comment": comm1, "result": True}


@pytest.mark.slow_test
def test_delete_beacon_module():
    """
    Test deleting a beacon.
    """
    comm1 = "Deleted beacon: watch_salt_master."
    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_delete_complete",
            "beacons": {},
        },
    ]

    with patch.dict(
        beacons.__opts__,
        {
            "beacons": {
                "watch_salt_master": [
                    {"processes": {"salt-master": "stopped"}},
                    {"beacon_module": "ps"},
                ]
            },
        },
    ):
        mock = MagicMock(return_value=True)
        with patch.dict(beacons.__salt__, {"event.fire": mock}):
            with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                assert beacons.delete("watch_salt_master") == {
                    "comment": comm1,
                    "result": True,
                }



"""
Manage python installations with pyenv.

.. note::
    Git needs to be installed and available via PATH if pyenv is to be
    installed automatically by the module.

.. versionadded:: 2014.4.0
"""

import logging
import os
import re
import shlex

# Set up logger
log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}

__opts__ = {
    "pyenv.root": None,
    "pyenv.build_env": None,
}


def _pyenv_exec(command, args="", env=None, runas=None, ret=None):
    if not is_installed(runas):
        return False

    binary = _pyenv_bin(runas)
    path = _pyenv_path(runas)

    if env:
        env = " {}".format(env)
    env = env or ""

    binary = "env PYENV_ROOT={}{} {}".format(path, env, binary)

    result = __salt__["cmd.run_all"](
        "{} {} {}".format(binary, command, args), runas=runas
    )

    if isinstance(ret, dict):
        ret.update(result)
        return ret

    if result["retcode"] == 0:
        return result["stdout"]
    else:
        return False


def _pyenv_bin(runas=None):
    path = _pyenv_path(runas)
    return "{}/bin/pyenv".format(path)


def _pyenv_path(runas=None):
    path = None
    if runas in (None, "root"):
        path = __salt__["config.option"]("pyenv.root") or "/usr/local/pyenv"
    else:
        path = __salt__["config.option"]("pyenv.root") or "~{}/.pyenv".format(runas)

    return os.path.expanduser(path)


def _install_pyenv(path, runas=None):
    if os.path.isdir(path):
        return True

    return 0 == __salt__["cmd.retcode"](
        "git clone https://github.com/yyuu/pyenv.git {}".format(path), runas=runas
    )


def _update_pyenv(path, runas=None):
    if not os.path.isdir(path):
        return False

    return 0 == __salt__["cmd.retcode"](
        "cd {} && git pull".format(shlex.quote(path)), runas=runas
    )


def _update_python_build(path, runas=None):
    path = "{}/plugins/python-build".format(path)
    if not os.path.isdir(path):
        return False

    return 0 == __salt__["cmd.retcode"](
        "cd {} && git pull".format(shlex.quote(path)), runas=runas
    )


def install(runas=None, path=None):
    """
    Install pyenv systemwide

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.install
    """
    path = path or _pyenv_path(runas)
    path = os.path.expanduser(path)
    return _install_pyenv(path, runas)


def update(runas=None, path=None):
    """
    Updates the current versions of pyenv and python-Build

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.update
    """
    path = path or _pyenv_path(runas)
    path = os.path.expanduser(path)

    return _update_pyenv(path, runas)


def is_installed(runas=None):
    """
    Check if pyenv is installed.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.is_installed
    """
    return __salt__["cmd.has_exec"](_pyenv_bin(runas))


def install_python(python, runas=None):
    """
    Install a python implementation.

    python
        The version of python to install, should match one of the
        versions listed by pyenv.list

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.install_python 2.0.0-p0
    """
    python = re.sub(r"^python-", "", python)

    env = None
    env_list = []

    if __grains__["os"] in ("FreeBSD", "NetBSD", "OpenBSD"):
        env_list.append("MAKE=gmake")

    if __salt__["config.option"]("pyenv.build_env"):
        env_list.append(__salt__["config.option"]("pyenv.build_env"))

    if env_list:
        env = " ".join(env_list)

    ret = {}
    ret = _pyenv_exec("install", python, env=env, runas=runas, ret=ret)
    if ret["retcode"] == 0:
        rehash(runas=runas)
        return ret["stderr"]
    else:
        # Cleanup the failed installation so it doesn't list as installed
        uninstall_python(python, runas=runas)
        return False


def uninstall_python(python, runas=None):
    """
    Uninstall a python implementation.

    python
        The version of python to uninstall. Should match one of the versions
        listed by :mod:`pyenv.versions <salt.modules.pyenv.versions>`

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.uninstall_python 2.0.0-p0
    """
    python = re.sub(r"^python-", "", python)

    args = "--force {}".format(python)
    _pyenv_exec("uninstall", args, runas=runas)
    return True


def versions(runas=None):
    """
    List the installed versions of python.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.versions
    """
    ret = _pyenv_exec("versions", "--bare", runas=runas)
    return [] if ret is False else ret.splitlines()


def default(python=None, runas=None):
    """
    Returns or sets the currently defined default python.

    python=None
        The version to set as the default. Should match one of the versions
        listed by :mod:`pyenv.versions <salt.modules.pyenv.versions>`. Leave
        blank to return the current default.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.default
        salt '*' pyenv.default 2.0.0-p0
    """
    if python:
        _pyenv_exec("global", python, runas=runas)
        return True
    else:
        ret = _pyenv_exec("global", runas=runas)
        return "" if ret is False else ret.strip()


def list_(runas=None):
    """
    List the installable versions of python.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.list
    """
    ret = []
    output = _pyenv_exec("install", "--list", runas=runas)
    if output:
        for line in output.splitlines():
            if line == "Available versions:":
                continue
            ret.append(line.strip())
    return ret


def rehash(runas=None):
    """
    Run pyenv rehash to update the installed shims.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.rehash
    """
    _pyenv_exec("rehash", runas=runas)
    return True


def do(cmdline=None, runas=None):
    """
    Execute a python command with pyenv's shims from the user or the system.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.do 'gem list bundler'
        salt '*' pyenv.do 'gem list bundler' deploy
    """
    path = _pyenv_path(runas)
    cmd_split = cmdline.split()
    quoted_line = ""
    for cmd in cmd_split:
        quoted_line = quoted_line + " " + shlex.quote(cmd)
    result = __salt__["cmd.run_all"](
        "env PATH={}/shims:$PATH {}".format(shlex.quote(path), quoted_line),
        runas=runas,
        python_shell=True,
    )

    if result["retcode"] == 0:
        rehash(runas=runas)
        return result["stdout"]
    else:
        return False


def do_with_python(python, cmdline, runas=None):
    """
    Execute a python command with pyenv's shims using a specific python version.

    CLI Example:

    .. code-block:: bash

        salt '*' pyenv.do_with_python 2.0.0-p0 'gem list bundler'
        salt '*' pyenv.do_with_python 2.0.0-p0 'gem list bundler' deploy
    """
    if python:
        cmd = "PYENV_VERSION={} {}".format(python, cmdline)
    else:
        cmd = cmdline

    return do(cmd, runas=runas)



"""
Management of OpenStack Keystone Services
=========================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.keystoneng` for setup instructions

Example States

.. code-block:: yaml

    create service:
      keystone_service.present:
        - name: glance
        - type: image

    delete service:
      keystone_service.absent:
        - name: glance

    create service with optional params:
      keystone_service.present:
        - name: glance
        - type: image
        - enabled: False
        - description: 'OpenStack Image'
"""


__virtualname__ = "keystone_service"


def __virtual__():
    if "keystoneng.service_get" in __salt__:
        return __virtualname__
    return (
        False,
        "The keystoneng execution module failed to load: shade python module is not"
        " available",
    )


def present(name, auth=None, **kwargs):
    """
    Ensure an service exists and is up-to-date

    name
        Name of the group

    type
        Service type

    enabled
        Boolean to control if service is enabled

    description
        An arbitrary description of the service
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_clouds"](auth)

    service = __salt__["keystoneng.service_get"](name=name)

    if service is None:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Service will be created."
            return ret

        kwargs["name"] = name
        service = __salt__["keystoneng.service_create"](**kwargs)
        ret["changes"] = service
        ret["comment"] = "Created service"
        return ret

    changes = __salt__["keystoneng.compare_changes"](service, **kwargs)
    if changes:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = changes
            ret["comment"] = "Service will be updated."
            return ret

        kwargs["name"] = service
        __salt__["keystoneng.service_update"](**kwargs)
        ret["changes"].update(changes)
        ret["comment"] = "Updated service"

    return ret


def absent(name, auth=None):
    """
    Ensure service does not exist

    name
        Name of the service
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    __salt__["keystoneng.setup_clouds"](auth)

    service = __salt__["keystoneng.service_get"](name=name)

    if service:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"id": service.id}
            ret["comment"] = "Service will be deleted."
            return ret

        __salt__["keystoneng.service_delete"](name=service)
        ret["changes"]["id"] = service.id
        ret["comment"] = "Deleted service"

    return ret



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest
from saltfactories.utils import random_string

from salt.exceptions import CommandExecutionError
from tests.support.case import ModuleCase

# Create group name strings for tests
ADD_GROUP = random_string("RS-", lowercase=False)
DEL_GROUP = random_string("RS-", lowercase=False)
CHANGE_GROUP = random_string("RS-", lowercase=False)
ADD_USER = random_string("RS-", lowercase=False)
REP_USER_GROUP = random_string("RS-", lowercase=False)


@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.skip_unless_on_darwin
class MacGroupModuleTest(ModuleCase):
    """
    Integration tests for the mac_group module
    """

    def setUp(self):
        """
        Sets up test requirements
        """
        os_grain = self.run_function("grains.item", ["kernel"])
        if os_grain["kernel"] not in "Darwin":
            self.skipTest("Test not applicable to '{kernel}' kernel".format(**os_grain))

    @pytest.mark.slow_test
    def test_mac_group_add(self):
        """
        Tests the add group function
        """
        try:
            self.run_function("group.add", [ADD_GROUP, 3456])
            group_info = self.run_function("group.info", [ADD_GROUP])
            self.assertEqual(group_info["name"], ADD_GROUP)
        except CommandExecutionError:
            self.run_function("group.delete", [ADD_GROUP])
            raise

    @pytest.mark.slow_test
    def test_mac_group_delete(self):
        """
        Tests the delete group function
        """
        # Create a group to delete - If unsuccessful, skip the test
        if self.run_function("group.add", [DEL_GROUP, 4567]) is not True:
            self.run_function("group.delete", [DEL_GROUP])
            self.skipTest("Failed to create a group to delete")

        # Now try to delete the added group
        ret = self.run_function("group.delete", [DEL_GROUP])
        self.assertTrue(ret)

    @pytest.mark.slow_test
    def test_mac_group_chgid(self):
        """
        Tests changing the group id
        """
        # Create a group to delete - If unsuccessful, skip the test
        if self.run_function("group.add", [CHANGE_GROUP, 5678]) is not True:
            self.run_function("group.delete", [CHANGE_GROUP])
            self.skipTest("Failed to create a group to manipulate")

        try:
            self.run_function("group.chgid", [CHANGE_GROUP, 6789])
            group_info = self.run_function("group.info", [CHANGE_GROUP])
            self.assertEqual(group_info["gid"], 6789)
        except AssertionError:
            self.run_function("group.delete", [CHANGE_GROUP])
            raise

    @pytest.mark.slow_test
    def test_mac_adduser(self):
        """
        Tests adding user to the group
        """
        # Create a group to use for test - If unsuccessful, skip the test
        if self.run_function("group.add", [ADD_GROUP, 5678]) is not True:
            self.run_function("group.delete", [ADD_GROUP])
            self.skipTest("Failed to create a group to manipulate")

        try:
            self.run_function("group.adduser", [ADD_GROUP, ADD_USER])
            group_info = self.run_function("group.info", [ADD_GROUP])
            self.assertEqual(ADD_USER, "".join(group_info["members"]))
        except AssertionError:
            self.run_function("group.delete", [ADD_GROUP])
            raise

    @pytest.mark.slow_test
    def test_mac_deluser(self):
        """
        Test deleting user from a group
        """
        # Create a group to use for test - If unsuccessful, skip the test
        if (
            self.run_function("group.add", [ADD_GROUP, 5678])
            and self.run_function("group.adduser", [ADD_GROUP, ADD_USER]) is not True
        ):
            self.run_function("group.delete", [ADD_GROUP])
            self.skipTest("Failed to create a group to manipulate")

        delusr = self.run_function("group.deluser", [ADD_GROUP, ADD_USER])
        self.assertTrue(delusr)

        group_info = self.run_function("group.info", [ADD_GROUP])
        self.assertNotIn(ADD_USER, "".join(group_info["members"]))

    @pytest.mark.slow_test
    def test_mac_members(self):
        """
        Test replacing members of a group
        """
        if (
            self.run_function("group.add", [ADD_GROUP, 5678])
            and self.run_function("group.adduser", [ADD_GROUP, ADD_USER]) is not True
        ):
            self.run_function("group.delete", [ADD_GROUP])
            self.skipTest(
                "Failed to create the {} group or add user {} to group "
                "to manipulate".format(ADD_GROUP, ADD_USER)
            )

        rep_group_mem = self.run_function("group.members", [ADD_GROUP, REP_USER_GROUP])
        self.assertTrue(rep_group_mem)

        # ensure new user is added to group and previous user is removed
        group_info = self.run_function("group.info", [ADD_GROUP])
        self.assertIn(REP_USER_GROUP, str(group_info["members"]))
        self.assertNotIn(ADD_USER, str(group_info["members"]))

    @pytest.mark.slow_test
    def test_mac_getent(self):
        """
        Test returning info on all groups
        """
        if (
            self.run_function("group.add", [ADD_GROUP, 5678])
            and self.run_function("group.adduser", [ADD_GROUP, ADD_USER]) is not True
        ):
            self.run_function("group.delete", [ADD_GROUP])
            self.skipTest(
                "Failed to create the {} group or add user {} to group "
                "to manipulate".format(ADD_GROUP, ADD_USER)
            )

        getinfo = self.run_function("group.getent")
        self.assertTrue(getinfo)
        self.assertIn(ADD_GROUP, str(getinfo))
        self.assertIn(ADD_USER, str(getinfo))

    def tearDown(self):
        """
        Clean up after tests
        """
        # Delete ADD_GROUP
        add_info = self.run_function("group.info", [ADD_GROUP])
        if add_info:
            self.run_function("group.delete", [ADD_GROUP])

        # Delete DEL_GROUP if something failed
        del_info = self.run_function("group.info", [DEL_GROUP])
        if del_info:
            self.run_function("group.delete", [DEL_GROUP])

        # Delete CHANGE_GROUP
        change_info = self.run_function("group.info", [CHANGE_GROUP])
        if change_info:
            self.run_function("group.delete", [CHANGE_GROUP])



"""
Tests for the templates utils
"""
import os

import pytest


def test_issue_60083(
    salt_call_cli,
    tmp_path,
    base_env_state_tree_root_dir,
):
    """
    Validate that we can serialize pillars to json in states.
    Issue #60083
    """
    target_path = tmp_path / "issue-60083-target.txt"
    assert not os.path.exists(str(target_path))
    sls_name = "issue-60083"
    sls_contents = """
    {{ pillar["target-path"] }}:
      file.managed:
        - contents: |
            {{ pillar|json }}
    """
    sls_tempfile = pytest.helpers.temp_file(
        "{}.sls".format(sls_name), sls_contents, base_env_state_tree_root_dir
    )
    with sls_tempfile:  # , issue_50221_ext_pillar_tempfile:
        ret = salt_call_cli.run(
            "state.apply", sls_name, pillar={"target-path": str(target_path)}
        )
        assert ret.stdout.find("Jinja error") == -1
        assert ret.data
        keys = list(ret.data.keys())
        assert len(keys) == 1
        key = keys[0]
        assert ret.data[key]["changes"]["diff"] == "New file"


def test_issue_62372(
    salt_call_cli,
    tmp_path,
    base_env_state_tree_root_dir,
):
    """
    Validate that we can use the random_* filters
    Issue #62372
    """
    target_path = tmp_path / "issue-62372-target.txt"
    assert not os.path.exists(str(target_path))
    sls_name = "issue-62372"
    sls_contents = """
    {% set my_list = ["one", "two", "three", "four"] -%}
    {{ pillar["target-path"] }}:
      file.managed:
        - contents: |
            {{ my_list | random_sample(2, seed="static") }}
            {{ my_list | random_shuffle(seed="static") }}
    """
    sls_tempfile = pytest.helpers.temp_file(
        "{}.sls".format(sls_name), sls_contents, base_env_state_tree_root_dir
    )
    with sls_tempfile:
        ret = salt_call_cli.run(
            "state.apply", sls_name, pillar={"target-path": str(target_path)}
        )
        assert ret.stdout.find("Jinja error") == -1
        assert ret.data
        keys = list(ret.data.keys())
        assert len(keys) == 1
        key = keys[0]
        assert ret.data[key]["changes"]["diff"] == "New file"
